#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
#UXML: UXML structured metadata
"""

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     prjemian@gmail.com
# :copyright: (c) 2014-2020, Pete R. Jemian
#
# Distributed under the terms of the Creative Commons Attribution 4.0 International Public License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------


from lxml import etree
import os
import six

from .. import eznx
from ..plugin import AutoRegister
from ..plugin import ControlLineHandler
from ..utils import strip_first_word


DEFAULT_XML_ROOT_TAG = "UXML"
UXML_PROVIDES_ROOT_TAG = False
XML_SCHEMA = os.path.join(os.path.dirname(__file__), "uxml.xsd")


class UXML_Error(Exception):
    pass


@six.add_metaclass(AutoRegister)
class UXML_metadata(ControlLineHandler):

    """
    **#UXML** -- XML metadata in scan header

    IN-MEMORY REPRESENTATION

    * (SpecDataFileScan): **UXML** : XML document root

    HDF5/NeXus REPRESENTATION

    * various items below the *NXentry* parent group,
      as indicated in the UXML

    .. rubric:: Public methods

    .. autosummary::

          ~process

    .. rubric:: Internal methods

    .. autosummary::

          ~walk_xml_tree
          ~make_NeXus_links
          ~prune_dict
          ~dataset
          ~group
          ~hardlink

    """

    key = r"#UXML"
    scan_attributes_defined = ["UXML", "UXML_root"]
    unique_id = {}
    target_id = {}
    selector = None
    converters = dict(int=int, float=float, str=str)

    def process(self, text, scan, *args, **kws):
        """read #UXML lines from SPEC data file into ``scan.UXML``"""
        if not hasattr(scan, "UXML"):
            scan.UXML = []

        line = strip_first_word(text)
        scan.UXML.append(line)
        scan.addPostProcessor("UXML_metadata", self.postprocess)

    def postprocess(self, scan, *args, **kws):
        """
        convert the UXML text into an XML object (``scan.UXML_root``)

        :param SpecDataFileScan scan: data from a single SPEC scan
        """
        xml_text = "\n".join(scan.UXML)
        if UXML_PROVIDES_ROOT_TAG:
            root = etree.fromstring(xml_text)
            # read root_tag from supplied UXML lines
        else:
            # provide default root tag
            xml_text = (
                "<%s>\n" % DEFAULT_XML_ROOT_TAG
                + xml_text
                + "\n</%s>" % DEFAULT_XML_ROOT_TAG
            )
            root = etree.fromstring(xml_text)

        scan.UXML_root = root
        # validate against the schema
        xml_schema_tree = etree.parse(XML_SCHEMA)
        xml_schema = etree.XMLSchema(xml_schema_tree)

        if not xml_schema.validate(root):
            # XML file is not valid, let lxml report what is wrong as an exception
            # log = xmlschema.error_log    # access more details
            try:
                xml_schema.assertValid(root)  # basic exception report
            except etree.DocumentInvalid as exc:
                emsg = "UXML error: " + str(exc)
                # logger.warn(emsg)
                raise UXML_Error(emsg)

        scan.addH5writer("UXML_metadata", self.writer)

    def writer(self, nxentry, writer, scan, *args, **kws):
        """Describe how to store this data in an HDF5 NeXus file"""
        self.unique_id = {}
        self.target_id = {}
        self.selector = dict(
            dataset=self.dataset, group=self.group, hardlink=self.hardlink
        )

        # parse the XML and store
        self.walk_xml_tree(
            eznx.makeGroup(
                nxentry, "UXML", "NXnote", desc="UXML metadata"
            ),
            scan.UXML_root,
        )
        self.make_NeXus_links()

    def walk_xml_tree(self, h5parent, xml_node):
        """parse the XML node into HDF5 objects"""
        for item in xml_node:
            handler = self.selector[item.tag]
            handler(h5parent, item)

    def make_NeXus_links(self):
        """create all the hardlinks as directed"""
        for target_id in self.target_id.keys():
            if target_id in self.unique_id:
                target_group, target_name = self.target_id[target_id]
                source = self.unique_id[target_id]
                eznx.makeLink(target_group, source, target_name)

    def prune_dict(self, d, keys):
        """remove keys from dictionary d"""
        return {k: v for k, v in d.items() if k not in keys}

    def dataset(self, h5parent, xml_node):
        """HDF5/NeXus dataset specification"""
        attrs = dict(xml_node.attrib)
        nm = attrs.get("name")
        data_type = attrs.get("type", "str")
        unique_id = attrs.get("unique_id")
        attrs = self.prune_dict(attrs, "name type unique_id".split())

        if data_type in self.converters:
            converter = self.converters[data_type]
            value = converter(xml_node.text)
        else:
            emsg = "unexpected type='%s'" % data_type
            raise UXML_Error(emsg)

        ds = eznx.makeDataset(h5parent, nm, value, **attrs)

        if unique_id is not None:
            self.unique_id[unique_id] = ds

        return ds

    def group(self, h5parent, xml_node):
        """HDF5/NeXus group specification"""
        attrs = dict(xml_node.attrib)
        nm = attrs.get("name")
        NX_class = attrs.get("NX_class")
        unique_id = attrs.get("unique_id")
        attrs = self.prune_dict(attrs, "name NX_class unique_id".split())

        group = eznx.makeGroup(h5parent, nm, NX_class, **attrs)

        if unique_id is not None:
            self.unique_id[unique_id] = group

        self.walk_xml_tree(group, xml_node)

        return group

    def hardlink(self, h5parent, xml_node):
        """HDF5/NeXus hard link specification"""
        attrs = dict(xml_node.attrib)
        nm = attrs.get("name")
        target_id = attrs.get("target_id")

        if target_id is not None:
            self.target_id[target_id] = (h5parent, nm)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# -----------------------------------------------------------------------------
# :author:    John Hammonds
# :email:     JPHammonds@anl.gov
# -----------------------------------------------------------------------------

"""
SPEC data file control lines unique to the APS XPCS instrument
"""

import six

from ..eznx import makeGroup
from ..plugin import AutoRegister, ControlLineHandler
from ..spec import SpecDataFileHeader, SpecDataFileScan
from ..utils import strip_first_word


@six.add_metaclass(AutoRegister)
class XPCS_VA(ControlLineHandler):
    """**#VA**"""

    key = r"#VA\d+"
    scan_attributes_defined = ["VA"]

    def process(self, text, spec_obj, *args, **kws):
        subkey = text.split()[0].lstrip("#")
        if not hasattr(spec_obj, "VA"):
            spec_obj.VA = {}
        spec_obj.VA[subkey] = strip_first_word(text)
        if isinstance(spec_obj, SpecDataFileHeader):
            spec_obj.addH5writer(self.key, self.writer)

    def writer(self, h5parent, writer, scan, nxclass=None, *args, **kws):
        """ Describe how to write VA data"""
        desc = "XPCS VA parameters"
        group = makeGroup(h5parent, "VA", nxclass, description=desc)
        dd = {}
        for item, value in scan.VA.items():
            dd[item] = list(map(str, value.split()))
        writer.save_dict(group, dd)


@six.add_metaclass(AutoRegister)
class XPCS_VD(ControlLineHandler):
    """**#VD** """

    key = r"#VD\d+"
    scan_attributes_defined = ["VD"]

    def process(self, text, spec_obj, *args, **kws):
        subkey = text.split()[0].lstrip("#")
        if not hasattr(spec_obj, "VD"):
            spec_obj.VD = {}

        spec_obj.VD[subkey] = strip_first_word(text)
        if isinstance(spec_obj, SpecDataFileHeader):
            spec_obj.addH5writer(self.key, self.writer)

    def writer(self, h5parent, writer, scan, nxclass=None, *args, **kws):
        """Describe how to write VD data"""
        desc = "XPCS VD parameters"
        group = makeGroup(h5parent, "VD", nxclass, description=desc)
        dd = {}
        for item, value in scan.VD.items():
            dd[item] = list(map(str, value.split()))
        writer.save_dict(group, dd)


@six.add_metaclass(AutoRegister)
class XPCS_VE(ControlLineHandler):
    """**#VE** """

    key = r"#VE\d+"
    scan_attributes_defined = ["VE"]

    def process(self, text, spec_obj, *args, **kws):
        subkey = text.split()[0].lstrip("#")
        if not hasattr(spec_obj, "VE"):
            spec_obj.VE = {}

        spec_obj.VE[subkey] = strip_first_word(text)
        if isinstance(spec_obj, SpecDataFileHeader):
            spec_obj.addH5writer(self.key, self.writer)

    def writer(self, h5parent, writer, scan, nxclass=None, *args, **kws):
        """Describe how to write VE data"""
        desc = "XPCS VE parameters"
        group = makeGroup(h5parent, "VE", nxclass, description=desc)
        dd = {}
        for item, value in scan.VE.items():
            dd[item] = list(map(str, value.split()))
        writer.save_dict(group, dd)


@six.add_metaclass(AutoRegister)
class XPCS_XPCS(ControlLineHandler):
    """#XPCS"""

    key = r"#XPCS"
    scan_attributes_defined = ["XPCS"]

    def process(self, text, spec_obj, *args, **kws):
        if not hasattr(spec_obj, "XPCS"):
            spec_obj.XPCS = {}
        splitWord = strip_first_word(text).split()
        spec_obj.XPCS[splitWord[0]] = splitWord[1:]
        if isinstance(spec_obj, SpecDataFileScan):
            spec_obj.addH5writer(self.key, self.writer)

    def writer(self, h5parent, writer, scan, nxclass=None, *args, **kws):
        pass


@six.add_metaclass(AutoRegister)
class XPCS_CCD(ControlLineHandler):
    """#CCD"""

    key = r"#CCD"
    scan_attributes_defined = ["CCD"]

    def process(self, text, spec_obj, *args, **kws):
        if not hasattr(spec_obj, "CCD"):
            spec_obj.CCD = {}
        splitWord = strip_first_word(text).split()
        spec_obj.CCD[splitWord[0]] = splitWord[1:]
        if isinstance(spec_obj, SpecDataFileScan):
            spec_obj.addH5writer(self.key, self.writer)

    def writer(self, h5parent, writer, scan, nxclass=None, *args, **kws):
        pass



#!/usr/bin/env python

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     prjemian@gmail.com
# :copyright: (c) 2014-2020, Pete R. Jemian
#
# Distributed under the terms of the Creative Commons Attribution 4.0 International Public License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------

"""
Plot the data from scan N in a SPEC data file

.. autosummary::

    ~Selector
    ~ImageMaker
    ~LinePlotter
    ~MeshPlotter
    ~openSpecFile

Exceptions:

.. autosummary::

    ~NoDataToPlot
    ~NotPlottable
    ~ScanAborted
    ~UnexpectedObjectTypeError

"""

import os
import numpy

from . import charts
from . import spec  # read SPEC data files
from . import singletons
from . import utils


class UnexpectedObjectTypeError(RuntimeError):

    """Incorrect Python object type: programmer error."""


class ScanAborted(RuntimeWarning):

    """Scan aborted before all points acquired."""


class NotPlottable(ValueError):

    """No plottable data for this scan."""


class NoDataToPlot(ValueError):

    """No data found."""


ABORTED_ATTRIBUTE_TEXT = "_aborted_"


class Selector(singletons.Singleton):
    """
    associate SPEC scan macro names with image makers

    :image maker: subclass of :class:`ImageMaker`

    To include a custom image maker from outside this module,
    create the subclass and then add it to an instance of this
    class.  Such as this plotter that defaults to a logarithmic
    scale for the X axis for all `logxscan` macros:

        from spec2nexus import specplot

        class LogX_Plotter(specplot.ImageMaker):

            def x_log(self):
                return True

        # ...

        selector = specplot.Selector()
        selector.add('logxscan', LogX_Plotter)

        # ...

        image_maker = specplot.Selector().auto(scan)
        plotter = image_maker()
        plotter.plot_scan(scan, fullPlotFile)

    This class is a singleton which means you will always get the same
    instance when you call this class many times in your program.

    .. autosummary::

        ~auto
        ~add
        ~update
        ~get
        ~exists
        ~default

    """

    default_key = "__default__"

    def __init__(self):
        self.db = {}
        self.add(self.default_key, LinePlotter, default=True)

    def auto(self, scan):
        """
        automatically choose a scan image maker based on the SPEC scan macro

        Selection Rules:

        * macro ends with "scan": use :class:`LinePlotter`
        * macro ends with "mesh": use :class:`MeshPlotter`
        * default: use default image maker (initially :class:`LinePlotter`)
        """
        if not isinstance(scan, (spec.SpecDataFileScan,)):
            msg = "expected a SPEC scan object, received: "
            msg += str(scan)
            raise UnexpectedObjectTypeError(msg)

        macro = scan.get_macro_name()
        if self.exists(macro):
            return self.get(macro)

        # adapt for different scan macros
        image_maker = self.default()
        if macro == "hklscan":
            image_maker = HKLScanPlotter
        elif macro.lower().endswith("scan"):
            image_maker = LinePlotter
        elif macro.lower().endswith("mesh"):
            image_maker = MeshPlotter

        # register this macro name
        self.add(macro, image_maker)
        return image_maker

    def add(self, key, value, default=False):
        """
        register a new value by key

        :param str key: name of key, typically the macro name
        :raises KeyError: if key exists
        :raises UnexpectedObjectTypeError: if value is not subclass of :class:`ImageMaker`
        """
        if self.exists(key):
            raise KeyError("key exists: " + key)
        if not issubclass(value, ImageMaker):
            msg = "expected subclass of ImageMaker, received type: "
            msg += type(value).__name__
            raise UnexpectedObjectTypeError(msg)

        self.db[key] = value
        if default:
            self.db[self.default_key] = value

        return value

    def update(self, key, value, default=False):
        """
        replace an existing key with a new value

        :param str key: name of key, typically the macro name
        :raises KeyError: if key does not exist
        :raises UnexpectedObjectTypeError: if value is not subclass of :class:`ImageMaker`
        """
        if not self.exists(key):
            raise KeyError("key does not exist: " + key)
        if not issubclass(value, ImageMaker):
            msg = "expected subclass of ImageMaker, received type: "
            msg += type(value).__name__
            raise UnexpectedObjectTypeError(msg)

        self.db[key] = value
        if default:
            self.db[self.default_key] = value

        return value

    def get(self, key):
        """
        return a value by key

        :returns: subclass of :class:`ImageMaker` or `None` if key not found
        """
        return self.db.get(key)

    def exists(self, key):
        """
        is the key known?
        """
        return key in self.db

    def default(self):
        """
        retrieve the value of the default key
        """
        return self.get(self.default_key)


class ImageMaker(object):
    """
    superclass to handle plotting of data from a SPEC scan

    .. rubric:: Internal data model

    :signal: name of the ``signal`` data (default data to be plotted)
    :data: values of various collected arrays {label: array}
    :axes: names of the axes of signal data

    .. rubric:: USAGE:

    #. Create a subclass of :class:`ImageMaker`
    #. Override any of these methods:

       .. autosummary::

            ~data_file_name
            ~make_image
            ~plottable
            ~plot_options
            ~retrieve_plot_data

    .. rubric:: EXAMPLE

    ::

        class LinePlotter(ImageMaker):
            '''create a line plot'''

            def make_image(self, plotFile):
                '''
                make MatPlotLib chart image from the SPEC scan

                :param obj plotData: object returned from :meth:`retrieve_plot_data`
                :param str plotFile: name of image file to write
                '''
                assert(self.signal in self.data)
                assert(len(self.axes) == 1)
                assert(self.axes[0] in self.data)

                y = self.data[self.signal]
                x = self.data[self.axes[0]]
                xy_plot(x, y,  plotFile,
                       title = self.plot_title(),
                       plot_subtitle = self.plot_subtitle(),
                       xtitle = self.x_title(),
                       ytitle = self.y_title(),
                       xlog = self.x_log(),
                       ylog = self.y_log(),
                       timestamp_str = self.timestamp())

        sfile = specplot.openSpecFile(specFile)
        scan = sfile.getScan(scan_number)
        plotter = LinePlotter()
        plotter.plot_scan(scan, plotFile, y_log=True)

    """

    def __init__(self):
        self.scan = None
        self.settings = self._initialize_settings_()
        self.signal = None
        self.axes = []
        self.data = {}

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # support methods that a subclass might override

    def data_file_name(self):
        """
        the name of the file with the actual data

        Usually, this is the SPEC data file
        but it *could* be something else
        """
        return self.scan.header.parent.fileName  # self.scan.specFile

    def make_image(self, plotFile):
        """
        make MatPlotLib chart image from the SPEC scan

        The data to be plotted are provided in:

        * `self.signal`
        * `self.axes`
        * `self.data`

        :param str plotFile: name of image file to write
        """
        raise NotImplementedError(
            "must implement make_image() in each subclass"
        )

    def plottable(self):
        """
        can this data be plotted as expected?
        """
        return False  # override in subclass with specific tests

    def plot_options(self):
        """
        re-define any plot options in a subclass
        """
        pass

    def retrieve_plot_data(self):
        """
        retrieve default plottable data from spec data file and store locally

        This method must retrieve the data to be plotted, either from the
        SPEC data file scan or from a file which name is provided
        in the scan detalis.

        These attributes must be set by this method:

        :data: dictionary containing values of the various collected arrays {label: array}
        :signal: name of the 'signal' data (default data to be plotted)
        :axes: names of the axes of signal data

        .. rubric:: Example data

        ::

            self.data = {
                'angle': [1, 2, 3, 4, 5],
                'counts': [0. 2. 55. 3. 0]}
            self.signal = 'counts'
            self.axes = ['angle']

        Raise any of these exceptions as appropriate:

        .. autosummary::

            ~NoDataToPlot
            ~NotPlottable
            ~ScanAborted

        """
        raise NotImplementedError(
            "must implement retrieve_plot_data() in each subclass"
        )

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # support methods that will not need to be defined in a subclass

    def data_is_newer_than_plot(self, plotFile):
        """only proceed if mtime of SPEC data file is newer than plotFile"""
        mtime_sdf = os.path.getmtime(self.data_file_name())
        if os.path.exists(plotFile):
            mtime_pf = os.path.getmtime(plotFile)
        else:
            mtime_pf = 0

        return mtime_sdf > mtime_pf

    def plot_scan(self, scan, plotFile, maker=None):
        """
        make an image plot of the data in the scan

        :param obj scan: instance of :class:`~spec2nexus.spec.SpecDataFileScan`
        :param str plotFile: file name for plot output
        """
        if not isinstance(scan, (spec.SpecDataFileScan,)):
            raise UnexpectedObjectTypeError(
                "scan object not a SpecDataFileScan"
            )
        if hasattr(scan, ABORTED_ATTRIBUTE_TEXT):
            match_text = "Scan aborted after 0 points."
            if scan.__getattribute__(ABORTED_ATTRIBUTE_TEXT) == match_text:
                raise ScanAborted(match_text)

        self.scan = scan

        self.set_plot_title(self.plot_title() or self.data_file_name())
        self.set_plot_subtitle(
            self.plot_subtitle()
            or "#" + str(self.scan.scanNum) + ": " + self.scan.scanCmd
        )
        self.set_timestamp(self.timestamp() or self.scan.date)

        try:
            self.retrieve_plot_data()
        except KeyError as _exc:
            if hasattr(self.scan, ABORTED_ATTRIBUTE_TEXT):
                raise ScanAborted(
                    self.scan.__getattribute__(ABORTED_ATTRIBUTE_TEXT)
                )
            raise _exc

        self.plot_options()

        if self.plottable() and self.data_is_newer_than_plot(plotFile):
            self.make_image(plotFile)

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # support the self.settings dictionary with get & set methods

    def _initialize_settings_(self):
        """
        initial values are set to `None`

        subclasses that set a value should first check if the value has already been set
        unless explicitly replacing any customizations by the user
        """
        return dict(
            title=None,
            subtitle=None,
            x_title=None,
            y_title=None,
            x_log=None,
            y_log=None,
            z_log=None,
            timestamp=None,
        )

    # TODO: apply property and setter decorations

    def plot_title(self):
        """Return the plot title."""
        return self.settings["title"]

    def set_plot_title(self, text):
        """Set the plot title."""
        self.settings["title"] = text

    def plot_subtitle(self):
        """Return the plot_subtitle."""
        return self.settings["subtitle"]

    def set_plot_subtitle(self, text):
        """Set the plot_subtitle."""
        self.settings["subtitle"] = text

    def x_title(self):
        """Return the title for the X axis."""
        return self.settings["x_title"]

    def set_x_title(self, text):
        """Set the x axis title."""
        self.settings["x_title"] = text

    def y_title(self):
        """Return the title for the Y axis."""
        return self.settings["y_title"]

    def set_y_title(self, text):
        """Set the y axis title."""
        self.settings["y_title"] = text

    def x_log(self):
        """Boolean: should the X axis be plotted on a log scale?"""
        return self.settings["x_log"]

    def set_x_log(self, choice):
        """Set the x axis logarithmic if True."""
        self.settings["x_log"] = choice

    def y_log(self):
        """Boolean: should the Y axis be plotted on a log scale?"""
        return self.settings["y_log"]

    def set_y_log(self, choice):
        """Set the y axis logarithmic if True."""
        self.settings["y_log"] = choice

    def z_log(self):
        """Boolean: should the Z axis (image) be plotted on a log scale?"""
        return self.settings["z_log"]

    def set_z_log(self, choice):
        """Set the z axis (image) logarithmic if True."""
        self.settings["z_log"] = choice

    def timestamp(self):
        """Return the time of this scan as a string."""
        return self.settings["timestamp"]

    def set_timestamp(self, text):
        """Set the plot time stamp."""
        self.settings["timestamp"] = text


class LinePlotter(ImageMaker):
    """
    create a line plot
    """

    def make_image(self, plotFile):
        """
        make MatPlotLib chart image from the SPEC scan

        :param str plotFile: name of image file to write
        """
        assert self.signal in self.data
        assert len(self.axes) == 1
        assert self.axes[0] in self.data

        y = self.data[self.signal]
        x = self.data[self.axes[0]]
        ts = self.timestamp()

        charts.xy_plot(
            x,
            y,
            plotFile,
            title=self.plot_title(),
            subtitle=self.plot_subtitle(),
            xtitle=self.x_title(),
            ytitle=self.y_title(),
            xlog=self.x_log(),
            ylog=self.y_log(),
            timestamp_str=ts,
        )

    def plottable(self):
        """
        can this data be plotted as expected?
        """
        if self.signal in self.data:
            signal = self.data[self.signal]
            if (
                signal is not None
                and len(signal) > 0
                and len(self.axes) == 1
            ):
                if len(signal) == len(self.data[self.axes[0]]):
                    return True
        return False

    def plot_options(self):
        """
        define the settings for this, accepting any non-default values first
        """
        self.x_title() or self.set_x_title(self.axes[0])
        self.y_title() or self.set_y_title(self.signal)
        self.x_log() or self.set_x_log(False)
        self.y_log() or self.set_y_log(False)
        self.set_z_log(False)

    def retrieve_plot_data(self):
        """retrieve default data from spec data file"""
        # plot last column v. first column
        assert isinstance(self.scan, spec.SpecDataFileScan)
        self.signal = self.scan.column_last
        if self.signal not in self.scan.data:
            raise NoDataToPlot(str(self.scan))
        self.axes = [
            self.scan.column_first,
        ]
        self.data = {
            label: self.scan.data.get(label)
            for label in self.scan.L
            if label in self.scan.data
        }


class HKLScanPlotter(LinePlotter):
    """
    create a line plot from hklscan macros
    """

    def retrieve_plot_data(self):
        """retrieve default data from spec data file"""
        # standard hklscan macro handling
        # find the real scan axis, the one that changes
        for axis in "H K L".split():
            data = self.scan.data.get(axis)
            if data is None:
                continue
            # could compare start & end from scanCmd, this looks simpler
            if min(data) != max(data):
                # tell it to use this axis instead
                self.axes = [
                    axis,
                ]
                break

        # if not found, default changes nothing
        if data is None:
            raise NotPlottable("no data in scan: " + str(self.scan))

        if len(self.axes) == 0:
            # issue #99:  file: lmn40.spe, scan 244, hkl all fixed
            axis = "data point number"
            data = range(1, 1 + len(self.scan.data["H"]))
            self.scan.data[axis] = data
            self.set_x_title(axis + " (hkl all held constant)")
            self.axes = [
                axis,
            ]

        self.scan.column_first = axis
        self.data[axis] = data
        self.signal = self.scan.column_last
        self.data[self.signal] = self.scan.data[self.signal]


class MeshPlotter(ImageMaker):
    """
    create a mesh plot (2-D image)

    ..rubric:: References:

    :mesh 2-D parser: http://www.certif.com/spec_help/mesh.html

        ::

            mesh motor1 start1 end1 intervals1 motor2 start2 end2 intervals2 time

    :hklmesh 2-D parser: http://www.certif.com/spec_help/hklmesh.html

        ::

            hklmesh Q1 start1 end1 intervals1 Q2 start2 end2 intervals2 time

    """

    # see code in: writer.Writer.mesh()        self._mesh_(scan)

    def make_image(self, plotFile):
        """
        make MatPlotLib chart image from the SPEC scan

        :param str plotFile: name of image file to write
        """
        if len(self.axes) == 2:
            image = self.data[self.signal]
            self.set_plot_subtitle(
                "%s,  %s" % (self.signal, self.scan.raw.splitlines()[0])
            )
            self.set_x_title(self.axes[0])
            self.set_y_title(self.axes[1])

            charts.make_png(
                image,
                plotFile,
                [self.data[axis] for axis in self.axes],
                title=self.plot_title(),
                subtitle=self.plot_subtitle(),
                timestamp_str=self.timestamp(),
                xtitle=self.x_title(),
                ytitle=self.y_title(),
                log_image=self.z_log(),
            )
        elif len(self.axes) == 1:
            # fallback to 1-D plot
            y = self.data[self.signal]
            x = self.data[self.axes[0]]
            charts.xy_plot(
                x,
                y,
                plotFile,
                title=self.plot_title(),
                subtitle=self.plot_subtitle(),
                xtitle=self.x_title(),
                ytitle=self.y_title(),
                xlog=self.x_log(),
                ylog=self.y_log(),
                timestamp_str=self.timestamp(),
            )

    def plottable(self):
        """
        can this data be plotted as expected?
        """
        try:
            assert self.signal in self.data
            signal = numpy.array(self.data[self.signal])
            assert len(self.axes) in (0, len(signal.shape))
            for order, axis in enumerate(reversed(self.axes)):
                assert axis in self.data
                assert signal.shape[order] == len(self.data[axis])
        except Exception:
            return False
        return True

    def plot_options(self):
        """
        define the settings for this, accepting any non-default values first
        """
        if len(self.axes) == 1:
            self.x_title() or self.set_x_title(self.axes[0])
            self.y_title() or self.set_y_title(self.signal)
        elif len(self.axes) == 2:
            self.x_title() or self.set_x_title(self.axes[1])
            self.y_title() or self.set_y_title(self.axes[0])
        self.x_log() or self.set_x_log(False)
        self.y_log() or self.set_y_log(False)
        self.z_log() or self.set_z_log(False)

    def retrieve_plot_data(self):
        """retrieve default data from spec data file

        data parser for 2-D mesh and hklmesh
        """
        (
            label1,
            _start1,
            _end1,
            intervals1,
            label2,
            _start2,
            _end2,
            intervals2,
            _time,
        ) = self.scan.scanCmd.split()[1:]
        if label1 not in self.scan.data:
            label1 = self.scan.L[0]  # mnemonic v. name
        if label2 not in self.scan.data:
            label2 = self.scan.L[1]  # mnemonic v. name
        axis1 = self.scan.data.get(label1)
        axis2 = self.scan.data.get(label2)
        intervals1, intervals2 = map(int, (intervals1, intervals2))
        # unused: start1, end1, start2, end2, time = map(float, (start1, end1, start2, end2, time))

        if len(axis1) < intervals1 and min(axis2) == max(axis2):
            # stopped scan before second row started, 1-D plot is better (issue #82)
            self.axes = [
                label1,
            ]
            self.signal = self.scan.column_last
            self.data[label1] = self.scan.data[label1]
            self.data[self.signal] = self.scan.data[self.signal]
            return

        axis1 = axis1[0 : intervals1 + 1]
        self.data[label1] = axis1  # 1-D array

        axis2 = [
            axis2[row]
            for row in range(len(axis2))
            if row % (intervals1 + 1) == 0
        ]
        self.data[label2] = axis2  # 1-D array

        column_labels = self.scan.L
        column_labels.remove(label1)  # special handling
        column_labels.remove(label2)  # special handling
        if self.scan.scanCmd.startswith("hkl"):
            # find the reciprocal space axis held constant
            label3 = [
                key for key in ("H", "K", "L") if key in column_labels
            ][0]
            self.data[label3] = self.scan.data.get(label3)[0]  # constant

        # build 2-D data objects (do not build label1, label2, [or label3] as 2-D objects)
        data_shape = [len(axis2), len(axis1)]
        for label in column_labels:
            if label not in self.data:
                axis = numpy.array(self.scan.data.get(label))
                self.data[label] = utils.reshape_data(axis, data_shape)
            else:
                pass

        self.signal = utils.clean_name(self.scan.column_last)
        self.axes = [label1, label2]

        if spec.MCA_DATA_KEY in self.scan.data:  # 3-D array(s)
            # save each spectrum
            for key, spectrum in sorted(
                self.scan.data[spec.MCA_DATA_KEY].items()
            ):
                num_channels = len(spectrum[0])
                data_shape.append(num_channels)
                mca = numpy.array(spectrum)
                data = utils.reshape_data(mca, data_shape)
                channels = range(1, num_channels + 1)
                ds_name = "_" + key + "_"
                self.data[ds_name] = data
                self.data[ds_name + "channel_"] = channels


# class NeXusPlotter(ImageMaker):    # TODO: issue #92
#     """
#     create a plot from a NeXus HDF5 data file
#     """
#
#     def retrieve_plot_data(self):
#         """retrieve default data from spec data file"""
#         raise NotImplementedError(self.__class__.__name__ + '() is not ready')
#
#     def make_image(self, plotFile):
#         """
#         make image file from the SPEC scan
#
#         :param str plotFile: name of image file to write
#         """
#         raise NotImplementedError(self.__class__.__name__ + '() is not ready')


def openSpecFile(specFile):
    """
    convenience routine so that others do not have to `import spec2nexus.spec`
    """
    sd = spec.SpecDataFile(specFile)
    return sd


def main():
    import argparse

    doc = __doc__.strip().splitlines()[0]
    p = argparse.ArgumentParser(description=doc)
    p.add_argument("specFile", help="SPEC data file name")
    p.add_argument(
        "scan_number", help="scan number in SPEC file", type=str
    )
    p.add_argument("plotFile", help="output plot file name")
    args = p.parse_args()

    sfile = openSpecFile(args.specFile)
    scan = sfile.getScan(args.scan_number)
    image_maker = Selector().auto(scan)
    plotter = image_maker()
    plotter.plot_scan(scan, args.plotFile)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     prjemian@gmail.com
# :copyright: (c) 2014-2020, Pete R. Jemian
#
# Distributed under the terms of the Creative Commons Attribution 4.0 International Public License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------

"""
SPEC data file standard control lines

:see: SPEC manual, *Standard Data File Format*,
   http://www.certif.com/spec_manual/user_1_4_1.html

"""

from collections import OrderedDict
import datetime
import six
import time

from ..diffractometers import get_geometry_catalog, Diffractometer
from ..eznx import write_dataset, makeGroup, openGroup, makeLink
from ..plugin import AutoRegister, ControlLineHandler
from ..scanf import scanf
from ..spec import (
    SpecDataFileHeader,
    SpecDataFileScan,
    DuplicateSpecScanNumber,
    MCA_DATA_KEY,
)
from ..utils import (
    strip_first_word,
    iso8601,
    split_column_labels,
    clean_name,
)
from ..writer import CONTAINER_CLASS


SCAN_DATA_KEY = "scan_data"

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

# header block


@six.add_metaclass(AutoRegister)
class SPEC_File(ControlLineHandler):

    """
    **#F** -- original data file name (starts a file header block)

    Module :mod:`spec2nexus.spec` is responsible for handling this control line.

    IN-MEMORY REPRESENTATION

    * (SpecDataFile): **fileName**
    * (SpecDataFileHeader) : **file**

    HDF5/NeXus REPRESENTATION

    * file root-level attribute: **SPEC_file**
    """

    key = r"#F"
    # do NOT add F to scan_attributes_defined

    def process(self, text, spec_file_obj, *args, **kws):
        if not hasattr(spec_file_obj, "specFile"):
            spec_file_obj.specFile = None
        if spec_file_obj.specFile in (None, ""):
            spec_file_obj.specFile = strip_first_word(text)


@six.add_metaclass(AutoRegister)
class SPEC_Epoch(ControlLineHandler):

    """
    **#E** -- the UNIX epoch (seconds from 00:00 GMT 1/1/70)

    In SPEC data files, the ``#E`` control line indicates the
    start of a *header* block.

    IN-MEMORY REPRESENTATION

    * (SpecDataFileHeader): **epoch** *int*

    HDF5/NeXus REPRESENTATION

    * file root-level attribute: **SPEC_epoch** *int*
    """

    key = r"#E"
    # do NOT add E to scan_attributes_defined

    def process(self, buf, sdf_object, *args, **kws):
        matches = [
            h for h in sdf_object.headers if h.raw.strip() == buf.strip()
        ]
        if len(matches) > 0:
            # this header exists, nothing to do
            return
        header = SpecDataFileHeader(buf, parent=sdf_object)
        line = buf.splitlines()[0].strip()
        if line.find(".") > -1:
            header.epoch = float(strip_first_word(line))
        else:
            header.epoch = int(strip_first_word(line))
        sdf_object.headers.append(header)
        header.interpret()  # parse the full header


@six.add_metaclass(AutoRegister)
class SPEC_Date(ControlLineHandler):

    """
    **#D** -- date/time stamp

    IN-MEMORY REPRESENTATION

    * (SpecDataFileHeader): **date** *str*, ISO8601 format

    HDF5/NeXus REPRESENTATION

    * file root-level attribute: **SPEC_date** *str* (value for 1st header block is used)
    """

    key = r"#D"
    # do NOT add epoch and date to scan_attributes_defined

    def process(self, text, sdf_object, *args, **kws):
        text = strip_first_word(text)
        # if isinstance(sdf_object, SpecDataFileScan):
        #     sdf_object = sdf_object.header
        sp = text.split(" ")
        if len(sp) == 1:
            sdf_object.epoch = int(float(text))
            text = datetime.datetime.fromtimestamp(float(text))
            text = text.strftime("%c")
        sdf_object.date = text
        if not hasattr(sdf_object, "epoch"):
            # Mon Jul 08 13:35:50 2019
            spec_fmt = "%a %b %d %H:%M:%S %Y"
            ts = time.strptime(text, spec_fmt)
            sdf_object.epoch = time.mktime(ts)
        if isinstance(sdf_object, SpecDataFileScan):
            sdf_object.addH5writer(self.key, self.writer)
            if len(sdf_object.header.date.strip()) == 0:
                sdf_object.header.date = text
                sdf_object.header.epoch = sdf_object.epoch
            # TODO: what if header date is greater than this scan's date?

    def writer(self, h5parent, writer, sdf_object, *args, **kws):
        """Describe how to store this data in an HDF5 NeXus file"""
        write_dataset(h5parent, "date", iso8601(sdf_object.date))


@six.add_metaclass(AutoRegister)
class SPEC_Comment(ControlLineHandler):

    """
    **#C** -- any comment either in the scan header or somewhere in the scan

    IN-MEMORY REPRESENTATION

    * (SpecDataFileHeader): **comments**
    * (SpecDataFileScan):  **comments**

    HDF5/NeXus REPRESENTATION

    * file root-level attribute: **SPEC_comments** :
      string array of all comments from first header block
    * dataset named **comments** under */NXentry* group, such as */S1/comments* :
      string array of all comments from this scan block

    """

    key = r"#C"
    scan_attributes_defined = ["comments"]

    def process(self, text, scan, *args, **kws):
        scan.comments.append(strip_first_word(text))
        pos = text.find("Scan aborted after ")
        if pos > 0:
            scan._aborted_ = text[pos:]
        if isinstance(scan, SpecDataFileScan):
            scan.addH5writer(self.key, self.writer)

    def writer(self, h5parent, writer, scan, *args, **kws):
        """Describe how to store this data in an HDF5 NeXus file"""
        write_dataset(
            h5parent, "comments", "\n".join(list(map(str, scan.comments)))
        )


# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

# scan block


@six.add_metaclass(AutoRegister)
class SPEC_Scan(ControlLineHandler):

    """
    **#S** -- SPEC scan

    In SPEC data files, the ``#S`` control line indicates the
    start of a *scan* block.  Each scan will be written to a
    separate **NXentry** group in the HDF5 file.

    **NXentry**:
        "The top-level NeXus group which contains all the data
        and associated information that comprise a single measurement."

        -- http://download.nexusformat.org/doc/html/classes/base_classes/NXentry.html

    IN-MEMORY REPRESENTATION

    * (SpecDataFile):
    * (SpecDataFileHeader):

    HDF5/NeXus REPRESENTATION

    * */NXentry* group named 'S%d` scan_number at root
      level, such as */S1*

    """

    key = r"#S"
    # do NOT add S to scan_attributes_defined

    def process(self, part, sdf, *args, **kws):
        if len(sdf.headers) == 0:
            # make a header if none exists now
            raw = ""  # TODO: what default content to use?
            header = SpecDataFileHeader(raw, parent=sdf)
            sdf.headers.append(header)
        else:
            header = sdf.headers[-1]  # pick the most recent header

        matches = [
            (k, s)
            for k, s in sdf.scans.items()
            if s.raw.strip() == part.strip()
        ]

        if len(matches) > 0 and sdf.last_scan is not None:
            if sdf.getScan(sdf.last_scan).raw != part:
                return

        scan = SpecDataFileScan(header, part, parent=sdf)
        if sdf.last_scan is not None:
            # We know that `part` does not match any existing scan.
            # Do the first few lines match (#S and #D in particular)?
            # If so, then that scan has been updated with more data.
            def beginning(buf):
                "return a string with first few lines of buf"
                return "\n".join(
                    [
                        line
                        for line in buf.strip().splitlines()[:5]
                        if line.startswith("#")
                    ]
                )

            # TODO: replace beginning() with new algorithm
            # if new_scan.raw.startswith(last_scan.raw), that's it!

            if beginning(part) == beginning(
                sdf.getScan(sdf.last_scan).raw
            ):
                # remove the last scan
                del sdf.scans[sdf.last_scan]
                sdf.last_scan = None

        scan.S = strip_first_word(part.splitlines()[0].strip())
        scan.scanNum = scan.S.split()[0]
        scan.scanCmd = strip_first_word(scan.S)

        if scan.scanNum in sdf.scans:
            # Before raising an exception,
            #    Check for duplicate scan number and create alternate name
            #    write as "%d.%d" % (scan.scanNum, index+1)
            #    where index is the lowest integer in
            #    range(really big) that is not already in use.
            # really_big <= len(total number of scans in data file)
            # Will a non-integer scanNum break anything?  [note: It *has* caused troubles.]
            for i in range(len(scan.parent.scans)):
                new_scanNum = "%s.%d" % (scan.scanNum, i + 1)
                if new_scanNum not in sdf.scans:
                    scan.scanNum = new_scanNum
                    break
        scan.scanNum = str(scan.scanNum)
        if scan.scanNum in sdf.scans:
            msg = str(scan.scanNum) + " in " + sdf.fileName
            raise DuplicateSpecScanNumber(msg)
        sdf.scans[scan.scanNum] = scan


@six.add_metaclass(AutoRegister)
class SPEC_Geometry(ControlLineHandler):

    """
    **#G** -- diffractometer geometry (numbered rows: #G0, #G1, ...)

    IN-MEMORY REPRESENTATION

    * (SpecDataFileScan): **G**

    HDF5/NeXus REPRESENTATION

    * *NXnote* group named **G** in the *NXentry* group, such as */S1/G*

      * Datasets created from dictionary <scan>.G
        (indexed by number from the scan block, such as ``G0``, ``G1``, ...).
        Meaning of contents for each index are defined by geometry-specific
        SPEC diffractometer support.

    * *NXinstrument* & *NXsample* groups for interpreted information
    """

    key = r"#G\d+"
    scan_attributes_defined = ["G", "diffractometer"]

    def process(self, text, scan, *args, **kws):
        subkey = text.split()[0].lstrip("#")
        scan.G[subkey] = strip_first_word(text)
        scan.addPostProcessor("diffractometer geometry", self.postprocess)

    def postprocess(self, scan, *args, **kws):
        if len(scan.G) > 0:
            scan.addH5writer("diffractometer geometry", self.writer)

        dgc = get_geometry_catalog()
        geometry = dgc.match(scan)
        diffractometer = Diffractometer(geometry)
        if diffractometer is None:
            scan.geometry = {}
        else:
            diffractometer.parse(scan)
            scan.diffractometer = diffractometer

    def writer(self, h5parent, writer, scan, nxclass=None, *args, **kws):
        """Describe how to store this data in an HDF5 NeXus file"""
        # e.g.: SPECD/four.mac
        # http://certif.com/spec_manual/fourc_4_9.html
        desc = "SPEC geometry arrays, meanings defined by SPEC diffractometer support"
        group = makeGroup(h5parent, "G", nxclass, description=desc)
        dd = {}
        for item, value in scan.G.items():
            dd[item] = list(map(float, value.split()))
        writer.save_dict(group, dd)

        gpar = scan.diffractometer.geometry_parameters
        if len(gpar) > 0:
            nxinstrument = openGroup(
                h5parent, "instrument", "NXinstrument"
            )
            write_dataset(
                nxinstrument,
                "name",
                scan.diffractometer.geometry_name_full,
            )
            if scan.diffractometer.lattice is not None:
                nxsample = openGroup(h5parent, "sample", "NXsample")
                abc = [
                    scan.diffractometer.lattice.a,
                    scan.diffractometer.lattice.b,
                    scan.diffractometer.lattice.c,
                ]
                angles = [
                    scan.diffractometer.lattice.alpha,
                    scan.diffractometer.lattice.beta,
                    scan.diffractometer.lattice.gamma,
                ]
                write_dataset(
                    nxsample, "unit_cell_abc", abc, units="angstrom",
                )
                write_dataset(
                    nxsample,
                    "unit_cell_alphabetagamma",
                    angles,
                    units="degrees",
                )
                write_dataset(  # ah, NeXus ... so many ways ...
                    nxsample, "unit_cell", abc + angles,
                )
            if "ub_matrix" in gpar:
                nxsample = openGroup(h5parent, "sample", "NXsample")
                ub = gpar["ub_matrix"].value
                write_dataset(nxsample, "ub_matrix", ub)
            if scan.diffractometer.mode is not None:
                nxsample = openGroup(h5parent, "sample", "NXsample")
                write_dataset(
                    nxsample,
                    "diffractometer_mode",
                    scan.diffractometer.mode,
                )
            if scan.diffractometer.sector is not None:
                nxsample = openGroup(h5parent, "sample", "NXsample")
                write_dataset(
                    nxsample,
                    "diffractometer_sector",
                    scan.diffractometer.sector,
                )
            if len(scan.diffractometer.reflections) > 0:
                nxsample = openGroup(h5parent, "sample", "NXsample")
                for i, ref in enumerate(scan.diffractometer.reflections):
                    nm = "or%d" % i
                    nxnote = openGroup(
                        nxsample,
                        nm,
                        "NXnote",
                        description=nm + ": orientation reflection",
                    )
                    write_dataset(nxnote, "h", ref.h)
                    write_dataset(nxnote, "k", ref.k)
                    write_dataset(nxnote, "l", ref.l)
                    write_dataset(
                        nxnote,
                        "wavelength",
                        ref.wavelength,
                        units="Angstrom",
                    )
                    for k, a in ref.angles.items():
                        write_dataset(
                            nxnote,
                            k,
                            a,
                            units="degrees",
                            description="diffractometer angle",
                        )
            if scan.diffractometer.wavelength is not None:
                # see: http://download.nexusformat.org/doc/html/strategies.html#strategies-wavelength
                nxmono = openGroup(
                    nxinstrument, "monochromator", "NXmonochromator"
                )
                ds = write_dataset(
                    nxmono,
                    "wavelength",
                    scan.diffractometer.wavelength,
                    units="angstrom",
                )
                # and: http://download.nexusformat.org/doc/html/classes/base_classes/NXbeam.html#nxbeam
                nxsample = openGroup(h5parent, "sample", "NXsample")
                nxbeam = openGroup(nxsample, "beam", "NXbeam")
                # link them
                makeLink(
                    nxsample, ds, nxbeam.name + "/incident_wavelength"
                )

            nxnote = openGroup(
                nxinstrument,
                "geometry_parameters",
                nxclass,
                description="SPEC geometry arrays, interpreted",
            )
            for kdv in gpar.values():
                d = kdv.description
                if len(d) == 0:
                    write_dataset(nxnote, kdv.key, kdv.value)
                else:
                    write_dataset(
                        nxnote, kdv.key, kdv.value, description=d
                    )


@six.add_metaclass(AutoRegister)
class SPEC_NormalizingFactor(ControlLineHandler):

    """
    **#I** -- intensity normalizing factor

    IN-MEMORY REPRESENTATION

    * (SpecDataFileScan): **I**

    HDF5/NeXus REPRESENTATION

    * Dataset named **intensity_factor** in the *NXentry* group, such as */S1/intensity_factor*
    """

    key = r"#I"
    scan_attributes_defined = ["I"]

    def process(self, text, scan, *args, **kws):
        scan.I = float(strip_first_word(text))

    def writer(self, h5parent, writer, scan, nxclass=None, *args, **kws):
        """Describe how to store this data in an HDF5 NeXus file"""
        if hasattr(scan, "I"):
            writer.write_dataset(h5parent, "intensity_factor", scan.I)


@six.add_metaclass(AutoRegister)
class SPEC_CounterNames(ControlLineHandler):

    """
    **#J** -- names of counters (each separated by two spaces) (new with SPEC v6)

    IN-MEMORY REPRESENTATION

    * (SpecDataFileHeader): **J** : mnemonics

    HDF5/NeXus REPRESENTATION

    * *NXnote* group named **counter_cross_reference** in the *NXentry* group, such as */S1/counter_cross_reference*

      * datasets with names supplied as SPEC counter mnemonics, string values supplied as SPEC counter names

    """

    key = r"#J\d+"
    scan_attributes_defined = ["J"]

    def process(self, text, header, *args, **kws):
        if not hasattr(header, "J"):
            header.J = []
        header.J.append(strip_first_word(text).split())
        header.addPostProcessor(
            "counter cross-referencing", self.postprocess
        )

    def postprocess(self, header, *args, **kws):
        counter_xref_postprocessing(header)
        header.addH5writer("counter cross-referencing", self.writer)

    def writer(self, h5parent, writer, header, nxclass=None, *args, **kws):
        """Describe how to store this data in an HDF5 NeXus file"""
        if not hasattr(header, "counter_xref"):
            header.counter_xref = {}  # mnemonic:name
        desc = "cross-reference SPEC counter mnemonics and names"
        comment = "keys are SPEC counter mnemonics, values are SPEC counter names"
        if nxclass is None:
            nxclass = CONTAINER_CLASS
        group = makeGroup(
            h5parent,
            "counter_cross_reference",
            nxclass,
            description=desc,
            comment=comment,
        )
        for key, value in sorted(header.counter_xref.items()):
            write_dataset(group, key, value)


@six.add_metaclass(AutoRegister)
class SPEC_CounterMnemonics(ControlLineHandler):

    """
    **#j** -- mnemonics of counter  (new with SPEC v6)

    IN-MEMORY REPRESENTATION

    * (SpecDataFileHeader): **j** : mnemonics

    HDF5/NeXus REPRESENTATION

    * *NXnote* group named **counter_cross_reference** in the *NXentry* group, such as */S1/counter_cross_reference*

      * datasets with names supplied as SPEC counter mnemonics, string values supplied as SPEC counter names

    """

    key = r"#j\d+"
    scan_attributes_defined = ["j"]

    def process(self, text, header, *args, **kws):
        if not hasattr(header, "j"):
            header.j = []
        header.j.append(strip_first_word(text).split())
        header.addPostProcessor(
            "counter cross-referencing", self.postprocess
        )

    def postprocess(self, header, *args, **kws):
        counter_xref_postprocessing(header)
        header.addH5writer("counter cross-referencing", self.writer)

    def writer(self, h5parent, writer, header, nxclass=None, *args, **kws):
        """Describe how to store this data in an HDF5 NeXus file"""
        if not hasattr(header, "counter_xref"):
            header.counter_xref = {}  # mnemonic:name
        desc = "cross-reference SPEC counter mnemonics and names"
        comment = "keys are SPEC counter mnemonics, values are SPEC counter names"
        if nxclass is None:
            nxclass = CONTAINER_CLASS
        group = makeGroup(
            h5parent,
            "counter_cross_reference",
            nxclass,
            description=desc,
            comment=comment,
        )
        for key, value in sorted(header.counter_xref.items()):
            write_dataset(group, key, value)


def counter_xref_postprocessing(header):
    if not hasattr(header, "j") or not hasattr(header, "J"):
        return
    if not hasattr(header, "counter_xref"):
        header.counter_xref = {}  # mnemonic:name
    for row_number, mne_row in enumerate(header.j):
        name_row = header.J[row_number]
        for column_number, mne in enumerate(mne_row):
            header.counter_xref[mne] = name_row[column_number]


@six.add_metaclass(AutoRegister)
class SPEC_Labels(ControlLineHandler):

    """
    **#L** -- data column labels

    IN-MEMORY REPRESENTATION

    * (SpecDataFileScan): **L** : labels
    * (SpecDataFileScan): **data** : {labels: values}

    HDF5/NeXus REPRESENTATION

    * *NXdata* group named **data** in the *NXentry* group,
      such as */S1/data*
    * datasets with names supplied in **L**, array values
      collected in **data_lines**

    """

    key = r"#L"
    scan_attributes_defined = ["L", "column_first", "column_last"]

    def process(self, text, scan, *args, **kws):
        # Some folks use more than two spaces!  Use regular expression(re) module
        scan.L = split_column_labels(strip_first_word(text))

        if len(scan.L) == 1 and hasattr(scan, "N") and scan.N[0] > 1:
            # BUT: some folks only use a single-space as a separator!
            # perhaps #L was written with single-space separators.?
            # Unusual for scan to have only 1 column, but possible
            labels = strip_first_word(text).split()
            if len(labels) == scan.N[0]:
                scan.L = labels

        scan.column_first = scan.L[0]
        scan.column_last = scan.L[-1]


@six.add_metaclass(AutoRegister)
class SPEC_Monitor(ControlLineHandler):

    """
    **#M** -- counting against this constant monitor count (see #T)

    IN-MEMORY REPRESENTATION

    * (SpecDataFileScan): **M**

    HDF5/NeXus REPRESENTATION

    * Dataset named **M** in the *NXentry* group, such as */S1/M*
    * Dataset named **counting_basis** in the *NXentry* group with
      value *SPEC scan with constant monitor count*, such as */S1/counting_basis*

    """

    key = r"#M"
    scan_attributes_defined = ["M", "monitor_name"]

    def process(self, text, scan, *args, **kws):
        text = strip_first_word(text)
        pos = text.find(" ")
        scan.M = text[:pos]
        scan.monitor_name = text[pos:].strip().lstrip("(").rstrip(")")

        scan.addH5writer(self.key, self.writer)

    def writer(self, h5parent, writer, scan, *args, **kws):
        """Describe how to store this data in an HDF5 NeXus file"""
        desc = "SPEC scan with constant monitor count"
        write_dataset(h5parent, "counting_basis", desc)
        write_dataset(
            h5parent, "M", float(scan.M), units="counts", description=desc
        )


@six.add_metaclass(AutoRegister)
class SPEC_NumColumns(ControlLineHandler):

    """
    **#N** -- number of columns of data [ num2 sets per row ]

    IN-MEMORY REPRESENTATION

    * (SpecDataFileScan): **N** : [int]

    HDF5/NeXus REPRESENTATION

    * not written to file
    """

    key = r"#N"
    # TODO: Needs an example data file to test (issue #8)
    scan_attributes_defined = ["N"]

    def process(self, text, scan, *args, **kws):
        scan.N = list(map(int, strip_first_word(text).split()))


@six.add_metaclass(AutoRegister)
class SPEC_PositionerNames(ControlLineHandler):

    """
    **#O** -- positioner names (numbered rows: #O0, #O1, ...)

    IN-MEMORY REPRESENTATION

    * (SpecDataFileHeader) : **O** : label
    * (SpecDataFileScan): **positioner** : {label: value}

    HDF5/NeXus REPRESENTATION

    * *NXnote* group named **positioners** in the *NXentry* group, such as */S1/positioners*

      * datasets created from dictionary <scan>.positioner

    * *NXnote* group named **positioner_cross_reference** in the *NXentry* group, such as */S1/positioner_cross_reference*

      * datasets with names supplied as SPEC positioner mnemonics, string values supplied as SPEC positioner names

    """

    key = r"#O\d+"
    scan_attributes_defined = ["O"]

    def process(self, text, sdf_object, *args, **kws):
        if isinstance(sdf_object, SpecDataFileScan):
            sdf_object = sdf_object.header
        key = text.split()[0]
        if key == "#O0":
            sdf_object.O = []  # TODO: What if motor names are different?
        content = strip_first_word(text).strip()
        if content == "":
            content = []
        else:
            content = split_column_labels(content)
        sdf_object.O.append(content)


@six.add_metaclass(AutoRegister)
class SPEC_PositionerMnemonics(ControlLineHandler):

    """
    **#o** -- positioner mnemonics (new with SPEC v6)

    IN-MEMORY REPRESENTATION

    * (SpecDataFileHeader): **o** : mnemonics

    HDF5/NeXus REPRESENTATION

    * *NXnote* group named **positioner_cross_reference** in the *NXentry* group, such as */S1/positioner_cross_reference*

      * datasets with names supplied as SPEC positioner mnemonics, string values supplied as SPEC positioner names

    """

    key = r"#o\d+"
    scan_attributes_defined = ["o"]

    def process(self, text, header, *args, **kws):
        if not hasattr(header, "o"):
            header.o = []
        content = strip_first_word(text).strip()
        if content == "":
            content = []
        else:
            content = content.split()
        header.o.append(content)
        header.addPostProcessor(
            "positioner cross-referencing", self.postprocess
        )

    def postprocess(self, header, *args, **kws):
        positioner_xref_postprocessing(header)
        header.addH5writer("positioner cross-referencing", self.writer)

    def writer(self, h5parent, writer, header, nxclass=None, *args, **kws):
        """Describe how to store this data in an HDF5 NeXus file"""
        if not hasattr(header, "positioner_xref"):
            header.counter_xref = {}  # mnemonic:name
        desc = "cross-reference SPEC positioner mnemonics and names"
        comment = "keys are SPEC positioner mnemonics, values are SPEC positioner names"
        if nxclass is None:
            nxclass = CONTAINER_CLASS
        group = makeGroup(
            h5parent,
            "positioner_cross_reference",
            nxclass,
            description=desc,
            comment=comment,
        )
        for key, value in sorted(header.positioner_xref.items()):
            write_dataset(group, key, value)


def positioner_xref_postprocessing(header):
    if not hasattr(header, "o") or not hasattr(header, "O"):
        return
    if not hasattr(header, "positioner_xref"):
        header.positioner_xref = {}  # mnemonic:name
    for row_number, mne_row in enumerate(header.o):
        name_row = header.O[row_number]
        for column_number, mne in enumerate(mne_row):
            header.positioner_xref[mne] = name_row[column_number]


@six.add_metaclass(AutoRegister)
class SPEC_Positioners(ControlLineHandler):

    """
    **#P** -- positioner values at start of scan (numbered rows: #P0, #P1, ...)

    IN-MEMORY REPRESENTATION

    * (SpecDataFileHeader) : **O** : label
    * (SpecDataFileScan): **positioner** : {label: value}

    HDF5/NeXus REPRESENTATION

    * *NXnote* group named **positioners** in the *NXentry* group, such as */S1/positioners*

      * datasets created from dictionary <scan>.positioner

    """

    key = r"#P\d+"
    scan_attributes_defined = ["P", "positioner"]

    def process(self, text, scan, *args, **kws):
        if isinstance(scan, SpecDataFileHeader):
            scan = scan.getLatestScan()
        content = strip_first_word(text)
        if content == "":
            content = []
        else:
            content = content.split()
        scan.P.append(content)
        scan.addPostProcessor("motor_positions", self.postprocess)

    def postprocess(self, scan, *args, **kws):
        """
        interpret the motor positions from the scan header

        :param SpecDataFileScan scan: data from a single SPEC scan
        """
        scan.positioner = OrderedDict()
        for row, values in enumerate(scan.P):
            if row >= len(scan.header.O):
                scan.add_interpreter_comment(
                    "#P%d found without #O%d" % (row, row)
                )
                continue
            for col, val in enumerate(values):
                if col >= len(scan.header.O[row]):
                    scan.add_interpreter_comment(
                        "extra value in #P%d position %d, no matching label in #O%d"
                        % (row, col + 1, row)
                    )
                    continue
                mne = scan.header.O[row][col]
                scan.positioner[mne] = float(val)
        if len(scan.positioner) > 0:
            scan.addH5writer(self.key, self.writer)

    def writer(self, h5parent, writer, scan, nxclass=None, *args, **kws):
        """Describe how to store this data in an HDF5 NeXus file"""
        desc = "SPEC positioners (#P & #O lines)"
        group = makeGroup(
            h5parent, "positioners", nxclass, description=desc
        )
        # writer.save_dict(group, scan.positioner)
        for k, v in scan.positioner.items():
            safe_name = clean_name(k)
            pg = openGroup(group, safe_name, "NXpositioner")
            write_dataset(pg, "name", safe_name, spec_name=k)
            write_dataset(pg, "value", v, spec_name=k)
        nxinstrument = openGroup(h5parent, "instrument", "NXinstrument")
        makeLink(h5parent, group, nxinstrument.name + "/positioners")


@six.add_metaclass(AutoRegister)
class SPEC_HKL(ControlLineHandler):

    """
    **#Q** -- :math:`Q` (:math:`hkl`) at start of scan

    IN-MEMORY REPRESENTATION

    * (SpecDataFileScan): **Q**

    HDF5/NeXus REPRESENTATION

    * Dataset named **Q** in the *NXentry* group, such as */S1/M*
    """

    key = r"#Q"
    scan_attributes_defined = ["Q"]

    def process(self, text, scan, *args, **kws):
        s = strip_first_word(text)
        if len(s) > 0:
            scan.Q = list(map(float, s.split()))
            scan.addH5writer(self.key, self.writer)

    def writer(self, h5parent, writer, scan, *args, **kws):
        """Describe how to store this data in an HDF5 NeXus file"""
        desc = "hkl at start of scan"
        write_dataset(h5parent, "Q", scan.Q, description=desc)


@six.add_metaclass(AutoRegister)
class SPEC_CountTime(ControlLineHandler):

    """
    **#T** -- counting against this constant number of seconds (see #M)

    IN-MEMORY REPRESENTATION

    * (SpecDataFileScan): **T**

    HDF5/NeXus REPRESENTATION

    * Dataset named **T** in the *NXentry* group, such as */S1/T*
    * Dataset named **counting_basis** in the *NXentry* group with
      value *SPEC scan with constant counting time*, such as */S1/counting_basis*

    """

    key = r"#T"
    scan_attributes_defined = ["T", "time_name"]

    def process(self, text, scan, *args, **kws):
        text = strip_first_word(text)
        pos = text.find(" ")
        scan.T = text[:pos]
        scan.time_name = text[pos:].strip().lstrip("(").rstrip(")")

        scan.addH5writer(self.key, self.writer)

    def writer(self, h5parent, writer, scan, *args, **kws):
        """Describe how to store this data in an HDF5 NeXus file"""
        desc = "SPEC scan with constant counting time"
        write_dataset(h5parent, "counting_basis", desc)
        write_dataset(
            h5parent, "T", float(scan.T), units="s", description=desc
        )


@six.add_metaclass(AutoRegister)
class SPEC_UserReserved(ControlLineHandler):

    """
    **#U** -- Reserved for user

    IN-MEMORY REPRESENTATION

    * (SpecDataFileHeader): **U**, [*str*]
    * (SpecDataFileScan): **U**, [*str*]

    HDF5/NeXus REPRESENTATION

    * Within a group named **UserReserved** in the *NXentry* group:
      dataset(s) named **header_##** (from the SPEC data file header
      section) or **item_##** (from the SPEC data file scan section),
      such as */S1/UserReserved/header_1* and  */S1/UserReserved/item_5*
    """

    key = r"#U"
    scan_attributes_defined = ["U"]

    def process(self, text, sdf_object, *args, **kws):
        text = strip_first_word(text)
        if not hasattr(sdf_object, "U"):
            sdf_object.U = []
        sdf_object.U.append(text.strip())

        sdf_object.addH5writer(self.key, self.writer)

    def writer(self, h5parent, writer, sdf_object, *args, **kws):
        """Describe how to store this data in an HDF5 NeXus file"""
        desc = 'SPEC control line "#U: Reserved for user"'
        group = openGroup(
            h5parent, "UserReserved", "NXnote", description=desc
        )
        if isinstance(sdf_object, SpecDataFileHeader):
            tag = "header"
        else:
            tag = "item"
        for i, text in enumerate(sdf_object.U):
            key = "%s_%d" % (tag, i + 1)
            write_dataset(
                group, key, text, description="#U line %d" % (i + 1)
            )


@six.add_metaclass(AutoRegister)
class SPEC_TemperatureSetPoint(ControlLineHandler):

    """
    **#X** -- Temperature Set Point (desired temperature)

    The default declaration of the #X control line is written::

         def Fheader '_cols++;printf("#X %gKohm (%gC)\\n",TEMP_SP,DEGC_SP)'

    The supplied macro alters this slightly (replacing %g with %f)
    and uses the :meth:`spec2nexus.scanf.scanf` implementation with this format::

        fmt = "#X %fKohm (%fC)"

    Depending on the circumstances, this might be a good candidate to override
    with a custom *ControlLineHandler* that parses the data as written.
    If the conversion process fails for any reason in this implementation, the *#X* line is ignored.

    IN-MEMORY REPRESENTATION

    * (SpecDataFileScan): **TEMP_SP**
    * (SpecDataFileScan): **DEGC_SP**

    HDF5/NeXus REPRESENTATION

    * Dataset named **TEMP_SP** in the *NXentry* group, such as */S1/TEMP_SP*
    * Dataset named **DEGC_SP** in the *NXentry* group, such as */S1/DEGC_SP*
    """

    key = r"#X"
    scan_attributes_defined = ["TEMP_SP", "DEGC_SP"]

    def process(self, text, scan, *args, **kws):
        # Try a list of formats until one succeeds
        format_list = [
            "#X %fKohm (%fC)",
            # "#X %g %g",        # note: %g specifier is not available
            "#X %f %f",
        ]
        for fmt in format_list:
            result = scanf(fmt, text)
            if result is not None:
                scan.TEMP_SP, scan.DEGC_SP = result
                scan.addH5writer(self.key, self.writer)

    def writer(self, h5parent, writer, scan, nxclass=None, *args, **kws):
        """Describe how to store this data in an HDF5 NeXus file"""
        # consider putting this info under NXsample or NXentry/metadata
        if hasattr(scan, "TEMP_SP"):
            write_dataset(
                h5parent,
                "TEMP_SP",
                scan.TEMP_SP,
                description="temperature set point",
            )
        if hasattr(scan, "DEGC_SP"):
            write_dataset(
                h5parent,
                "DEGC_SP",
                scan.DEGC_SP,
                units="C",
                description="temperature set point (C)",
            )


@six.add_metaclass(AutoRegister)
class SPEC_DataLine(ControlLineHandler):

    """
    **(scan_data)** -- scan data line

    Scan data could include interspersed MCA data or
    even describe 2-D or 3-D data.  T
    his method reads the data lines and
    buffers them for post-processing in
    :meth:`spec2nexus.plugins.spec_common_spec2nexus.data_lines_postprocessing`.

    IN-MEMORY REPRESENTATION

    * (SpecDataFileScan): **data_lines** : values
    * (SpecDataFileScan): **data** : {labels: values}

    HDF5/NeXus REPRESENTATION

    * *NXdata* group named **data** in the *NXentry* group,
      such as */S1/data*

    * datasets with names supplied in **L**, array values
      collected in **data_lines**

    """

    # key = r'[+-]?\d*\.?\d?'
    # use custom key match since regexp for floats is tedious!
    key = SCAN_DATA_KEY
    scan_attributes_defined = ["data", "data_lines"]

    def match_key(self, text):
        """
        Easier to try conversion to number than construct complicated regexp
        """
        try:
            float(text.strip().split()[0])
            return True
        except ValueError:
            return False

    def process(self, text, scan, *args, **kws):
        scan.data_lines.append(text)

        # defer processing since comments and MCA data may
        # intersperse the scan data
        scan.addPostProcessor(SCAN_DATA_KEY, self.postprocess)

    def postprocess(self, scan, *args, **kws):
        data_lines_postprocessing(scan)


# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

# MCA: multi-channel analyzer

# see ESRF BLISS group: http://www.esrf.eu/blissdb/macros/getsource.py?macname=mca.mac


@six.add_metaclass(AutoRegister)
class SPEC_MCA(ControlLineHandler):

    """
    **#@MCA** -- MCA data formatting declaration (ignored for now)

    declares this scan contains MCA data (SPEC's array_dump() format, such as ``"#@MCA 16C"``)

    From documentation provided by the ESRF BLISS group:
    (http://www.esrf.eu/blissdb/macros/getsource.py?macname=mca.mac)

        #@MCA 16C
        Format string passed to data_dump() function.
        This format string is held by the global variable "MCA_FMT" and can then been adapted to particular needs.
        "%%16C" is the default. It dumps data on 1 line, cut every 16 points::

            @A 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0\\
             0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0\\
             0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0\\
             0 0 0 0 0 0 0 0 0 0 0 ...

        "%%16" would do the same without any backslash
        "1" would dump 1 point per line, ...

    """

    key = r"#@MCA"

    def process(self, text, scan, *args, **kws):
        # #@MCA 16C
        # Isn't this only informative to how the data is presented in the file?
        pass  # not sure how to handle this, ignore it for now


@six.add_metaclass(AutoRegister)
class SPEC_MCA_Array(ControlLineHandler):

    """
    **@A** -- MCA Array data

    MCA data. Each value is the content of one channel, or an
    integrated value over several channels if a reduction was applied.

    Since the MCA Array data is interspersed with scan data,
    this method reads the data lines and buffers them for post-processing in
    :meth:`spec2nexus.plugins.spec_common_spec2nexus.data_lines_postprocessing`.

    IN-MEMORY REPRESENTATION

    * (SpecDataFileScan): **data_lines** : values
    * (SpecDataFileScan): **data** : {labels: values}

    HDF5/NeXus REPRESENTATION

    * *NXdata* group named **data** in the *NXentry* group, such as */S1/data*

      * Dataset **_mca_** : *float* MCA data reported on *@A* lines
      * Dataset **_mca_channel_**: provided as HDF5 dimension scale for **_mca_** dataset
          * if CALIB data specified: *float* scaled MCA channels -- :math:`x_k = a +bk + ck^2`
          * if CALIB data not specified: *int* MCA channel numbers

    """

    key = r"@A\d*"
    # continued lines will be matched by SPEC_DataLine
    # process these lines only after all lines have been read
    scan_attributes_defined = ["data_lines"]

    # TODO: need more examples of MCA spectra in SPEC files to improve this
    # Are there any other MCA spectra (such as @B) possible?

    def process(self, text, scan, *args, **kws):
        # acquire like numerical data, handle in postprocessing
        scan.data_lines.append(text)
        scan.addPostProcessor(SCAN_DATA_KEY, self.postprocess)

    def postprocess(self, scan, *args, **kws):
        data_lines_postprocessing(scan)


@six.add_metaclass(AutoRegister)
class SPEC_MCA_Calibration(ControlLineHandler):

    """
    **#@CALIB** -- coefficients to compute a scale based on the MCA channel number

    :math:`x_k = a +bk + ck^2` for MCA data, :math:`k` is channel number

    IN-MEMORY REPRESENTATION

    * (SpecDataFileScan): **MCA['CALIB']** = ``dict(a, b, c)``

    HDF5/NeXus REPRESENTATION

    * defines a dimension scale for MCA data
    * *NXnote* group named **MCA** in the *NXentry* group, such as */S1/MCA*

      * Dataset **calib_a** : *float*
      * Dataset **calib_b** : *float*
      * Dataset **calib_c** : *float*

    """

    # key = r'#@CALIB'
    # accept upper or lower case variants
    # https://certif.com/spec_help/scans.html
    key = r"#@[cC][aA][lL][iI][bB]"
    scan_attributes_defined = ["MCA"]

    def process(self, text, scan, *args, **kws):
        # #@CALIB a b c
        # #@Calib 0.0501959 0.0141105 0 mca1
        s = strip_first_word(text).split()
        a, b, c = list(map(float, s[0:3]))

        if not hasattr(scan, "MCA"):
            scan.MCA = {}
        if "CALIB" not in scan.MCA:
            scan.MCA["CALIB"] = {}

        scan.MCA["CALIB"]["a"] = a
        scan.MCA["CALIB"]["b"] = b
        scan.MCA["CALIB"]["c"] = c
        scan.addH5writer(self.key, self.writer)

    def writer(self, h5parent, writer, scan, nxclass=None, *args, **kws):
        """Describe how to store this data in an HDF5 NeXus file"""
        if hasattr(scan, "MCA"):
            if "CALIB" in scan.MCA:
                mca_group = openGroup(
                    h5parent, "MCA", nxclass, description="MCA metadata"
                )
                calib_dict = scan.MCA["CALIB"]
                for key in "a b c".split():
                    if key in calib_dict:
                        write_dataset(
                            mca_group, "calib_" + key, calib_dict[key]
                        )


@six.add_metaclass(AutoRegister)
class SPEC_MCA_ChannelInformation(ControlLineHandler):

    """
    **#@CHANN** -- MCA channel information

    number_saved, first_saved, last_saved, reduction_coef

    IN-MEMORY REPRESENTATION

    * (SpecDataFileScan): **MCA['CALIB']** = ``dict(number_saved, first_saved, last_saved, reduction_coef)``

    HDF5/NeXus REPRESENTATION

    * *NXnote* group named **MCA** in the *NXentry* group, such as */S1/MCA*

      * Dataset **number_saved** : *int* number of channels saved
      * Dataset **first_saved** : *int* first channel saved
      * Dataset **first_saved** : *int* last channel saved
      * Dataset **reduction_coef** : *float* reduction coefficient

    """

    key = r"#@CHANN"
    scan_attributes_defined = ["MCA"]

    def process(self, text, scan, *args, **kws):
        # #@CHANN 1201 1110 1200 1
        s = strip_first_word(text).split()
        number_saved, first_saved, last_saved = list(map(int, s[0:3]))
        reduction_coef = float(s[-1])

        if not hasattr(scan, "MCA"):
            scan.MCA = {}

        scan.MCA["number_saved"] = number_saved
        scan.MCA["first_saved"] = first_saved
        scan.MCA["last_saved"] = last_saved
        scan.MCA["reduction_coef"] = reduction_coef
        scan.addH5writer(self.key, self.writer)

    def writer(self, h5parent, writer, scan, nxclass=None, *args, **kws):
        """Describe how to store this data in an HDF5 NeXus file"""
        if hasattr(scan, "MCA"):
            mca_group = openGroup(
                h5parent, "MCA", nxclass, description="MCA metadata"
            )
            mca = scan.MCA
            for (
                key
            ) in "number_saved  first_saved  last_saved  reduction_coef".split():
                if key in mca:
                    write_dataset(mca_group, key, mca[key])

            # make link in NXinstrument group
            nxinstrument = openGroup(
                h5parent, "instrument", "NXinstrument"
            )
            if "MCA" not in nxinstrument:
                makeLink(h5parent, mca_group, nxinstrument.name + "/MCA")


@six.add_metaclass(AutoRegister)
class SPEC_MCA_CountTime(ControlLineHandler):

    """
    **#@CTIME** -- MCA count times

    preset_time, elapsed_live_time, elapsed_real_time

    IN-MEMORY REPRESENTATION

    * (SpecDataFileScan): **MCA['CALIB']** = ``dict(preset_time, elapsed_live_time, elapsed_real_time)``

    HDF5/NeXus REPRESENTATION

    * *NXnote* group named **MCA** in the *NXentry* group, such as */S1/MCA*

      * Dataset **preset_time** : *float*
      * Dataset **elapsed_live_time** : *float*
      * Dataset **elapsed_real_time** : *float*

    """

    key = r"#@CTIME"
    scan_attributes_defined = ["MCA"]

    def process(self, text, scan, *args, **kws):
        s = strip_first_word(text).split()
        preset_time, elapsed_live_time, elapsed_real_time = list(
            map(float, s)
        )

        if not hasattr(scan, "MCA"):
            scan.MCA = {}

        scan.MCA["preset_time"] = preset_time
        scan.MCA["elapsed_live_time"] = elapsed_live_time
        scan.MCA["elapsed_real_time"] = elapsed_real_time
        scan.addH5writer(self.key, self.writer)

    def writer(self, h5parent, writer, scan, nxclass=None, *args, **kws):
        """Describe how to store this data in an HDF5 NeXus file"""
        if hasattr(scan, "MCA"):
            mca_group = openGroup(
                h5parent, "MCA", nxclass, description="MCA metadata"
            )
            mca = scan.MCA
            for (
                key
            ) in (
                "preset_time  elapsed_live_time  elapsed_real_time".split()
            ):
                if key in mca:
                    write_dataset(mca_group, key, mca[key], units="s")

            # make link in NXinstrument group
            nxinstrument = openGroup(
                h5parent, "instrument", "NXinstrument"
            )
            if "MCA" not in nxinstrument:
                makeLink(h5parent, mca_group, nxinstrument.name + "/MCA")


@six.add_metaclass(AutoRegister)
class SPEC_MCA_RegionOfInterest(ControlLineHandler):

    """
    **#@ROI** -- MCA ROI (Region Of Interest) channel information

    ROI_name, first_chan, last_chan

    IN-MEMORY REPRESENTATION

    * (SpecDataFileScan): **MCA['ROI']** = {ROI_name:dict(first_chan, last_chan)}

    HDF5/NeXus REPRESENTATION

    * *NXnote* group **ROI** in
      in *NXnote* group named **MCA**
      in the *NXentry* group, such as */S1/MCA/ROI*

      * Dataset **{ROI_name}** : *int* [first_chan, last_chan]

    """

    key = r"#@ROI"
    scan_attributes_defined = ["MCA"]

    def process(self, text, scan, *args, **kws):
        text = strip_first_word(text)

        pos = text.rfind(" ")
        last_chan = int(text[pos:])
        text = text[:pos]

        pos = text.rfind(" ")
        first_chan = int(text[pos:])
        ROI_name = text[:pos].strip()

        if not hasattr(scan, "MCA"):
            scan.MCA = {}
        if "ROI" not in scan.MCA:
            scan.MCA["ROI"] = {}

        scan.MCA["ROI"][ROI_name] = {}
        scan.MCA["ROI"][ROI_name]["first_chan"] = first_chan
        scan.MCA["ROI"][ROI_name]["last_chan"] = last_chan

    def writer(self, h5parent, writer, scan, nxclass=None, *args, **kws):
        """Describe how to store this data in an HDF5 NeXus file"""
        if hasattr(scan, "MCA"):
            if hasattr(scan.MCA, "ROI"):
                mca_group = openGroup(
                    h5parent, "MCA", nxclass, description="MCA metadata"
                )
                roi_group = openGroup(
                    mca_group,
                    "ROI",
                    nxclass,
                    description="Regions Of Interest",
                )
                roi_dict = scan.MCA["ROI"]
                for key, roi in roi_dict.items():
                    dataset = [roi["first_chan"], roi["last_chan"]]
                    desc = "first_chan, last_chan"
                    write_dataset(
                        roi_group,
                        key,
                        dataset,
                        description=desc,
                        units="channel",
                    )

            # make link in NXinstrument group
            nxinstrument = openGroup(
                h5parent, "instrument", "NXinstrument"
            )
            if "MCA" not in nxinstrument:
                makeLink(h5parent, mca_group, nxinstrument.name + "/MCA")


# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -


def combine_split_NM_lines(nm, data_lines):
    """
    combine split lines of data

        #N N [M]

    Indicates there are N columns of data.
    If M is present, it indicates there
    are M sets of data columns on each line.
    """
    if len(nm) == 1:
        # no split lines indicated
        return data_lines
    dl = []
    N, M = nm
    buf = ""
    for line in data_lines:
        if line.startswith("@A"):
            if len(buf) > 0:
                dl.append(buf)
            buf = []
            dl.append(line)
        else:
            if len(buf) > 0 and buf[-1] != " ":
                buf += " "  # ensure whitespace delimiter
            buf += line
            if len(buf.split()) == N:  # assumption here!
                dl.append(buf)
                buf = ""

    return dl


def data_lines_postprocessing(scan):
    """
    interpret the data lines from the body of the scan

    :param SpecDataFileScan scan: data from a single SPEC scan
    """
    if len(scan.L) == 1 and scan.L[0].split(" ") == scan.N[0]:
        # https://github.com/prjemian/spec2nexus/issues/216#issuecomment-569745297
        raise ValueError(
            "file {}: "
            "#S {}: "
            "#L lines must use two (2) spaces between labels".format(
                scan.specFile, scan.S
            )
        )

    if len(scan.L) != scan.N[0]:
        # https://github.com/prjemian/spec2nexus/issues/216
        raise ValueError(
            "file {}: "
            "#S {}: "
            "# of given column labels in #L ({}) "
            "does not match # specified in #N ({})".format(
                scan.specFile, scan.S, len(scan.L), scan.N[0]
            )
        )

    # first, get the column labels, rename redundant labels to be unique
    # the unique labels will be the scan.data dictionary keys
    scan.data = {}
    for col in range(len(scan.L)):
        label = scan._unique_key(scan.L[col], scan.data.keys())
        # need to guard when same column label is used more than once
        if label != scan.L[col]:
            scan.L[col] = label  # rename this column's label
        scan.data[label] = []  # list for the column's data
    num_columns = len(scan.data)

    # any MCA spectra?
    data_lines = scan.data_lines
    if "@A" in [line[:2] for line in data_lines if len(line) > 2]:
        # There can be more than 1 MCA spectrum specified
        # scan.data[MCA_DATA_KEY] = {}  keys: mca or mca1, mca2, ...
        scan.data[MCA_DATA_KEY] = {}

    data_lines = combine_split_NM_lines(scan.N, data_lines)

    # interpret the data lines from the body of the scan
    for values in data_lines:
        if values.startswith("@A"):
            # which MCA spectrum is THIS one?
            parts = values.split()
            if parts[0] == "@A":  # @A: mca
                key = r"mca"
            else:
                key = r"mca" + parts[0][2:]  # @A1: mca1, @A2: mca2, ...
            if key not in scan.data[MCA_DATA_KEY]:
                scan.data[MCA_DATA_KEY][key] = []
            # accumulate this spectrum
            mca_spectrum = list(map(float, parts[1:]))
            scan.data[MCA_DATA_KEY][key].append(mca_spectrum)
        else:
            try:
                buf = scan._interpret_data_row(values)
                if len(buf) == num_columns:
                    # only keep complete rows
                    for label, val in buf.items():
                        scan.data[label].append(val)
            except ValueError as _exc:
                pass  # ignore bad data lines (could save it as such ...)
    scan.addH5writer(SCAN_DATA_KEY, data_lines_writer)


def data_lines_writer(h5parent, writer, scan, *args, **kws):
    """Describe how to store scan data in an HDF5 NeXus file"""
    desc = "SPEC scan data"
    nxdata = makeGroup(h5parent, "data", "NXdata", description=desc)
    writer.save_data(nxdata, scan)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     prjemian@gmail.com
# :copyright: (c) 2014-2020, Pete R. Jemian
#
# Distributed under the terms of the Creative Commons Attribution 4.0 International Public License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------


"""
Provides a set of classes to read the contents of a SPEC data file.

:author: Pete Jemian
:email: jemian@anl.gov

:meth:`~spec2nexus.spec.SpecDataFile` is the only class users will need to call.
All other :mod:`~spec2nexus.spec` classes are called from this class.
The :meth:`~spec2nexus.spec.SpecDataFile.read` method is called automatically.

The user should create a class instance for each spec data file,
specifying the file reference (by path reference as needed)
and the internal routines will take care of all that is necessary
to read and interpret the information.

.. autosummary::

    ~is_spec_file
    ~is_spec_file_with_header
    ~SpecDataFile
    ~SpecDataFileHeader
    ~SpecDataFileScan


..  -----------------------------------------------------------------------------------------
    old documentation
    -----------------------------------------------------------------------------------------

    .. index:: SPEC data file structure

    The parser makes the assumption that a SPEC data file is composed from
    a sequence of component blocks.  The component blocks are either header
    or scan blocks.  Header blocks have the first line starting with ``#F``
    while scan blocks have the first line starting with ``#S``.  Usually,
    there is only one header block in a SPEC data file, followed by many
    scan blocks.  The header block contains information common to all the
    scan blocks that follow it.  Content for each block continues until
    the next block starts or the file ends.  The pattern is:

    * #F line starts a header block
    * there could be multiple #F lines in a data file
    * #S lines start a SPEC scan
    * everything between #F and the next #S is header content
    * everything after a #S line is scan content (until EOF, the next #S or the next #F)

    .. rubric:: Additional assumptions

    * Lines that begin with ``#`` contain metadata of some form.
    * Lines that begin with ``@`` contain MCA data
    * Lines that begin with a number are data points
    * Line that are blank will be ignored
    * Lines that begin with anything else are unexpected and will be ignored

    For lines that begin with ``#``, these hold keys to some form of metadata.
    Some of the keys are identified and used by the SPEC standard.mac (and other)
    macro files.  Other keys are left to the user to define.  There are two
    general types of key, best described by a regular expression:

    ====================  ============  ============================
    regexp                example       how it appears
    ====================  ============  ============================
    ``^#[a-zA-Z]+\s``     ``#S``        by itself
    ``^#[a-zA-Z]+\d+\s``  ``#P5``       part of a numbered series
    ====================  ============  ============================

    Note that keys that appear as part of a numbered series (such as ``#O0 #O1 #O2`` ...),
    usually have numbers starting at 0.
    -----------------------------------------------------------------------------------------

Note that the SPEC geometry control lines (``#G0 #G1`` ...)
have meanings that are unique to specific diffractometer geometries including
different numbers of values.  Consult the geometry macro file for specifics.



.. rubric:: Examples

Get the first and last scan numbers from the file:

    >>> from spec2nexus import spec
    >>> spec_data = spec.SpecDataFile('path/to/my/spec_data.dat')
    >>> print(spec_data.fileName)
    path/to/my/spec_data.dat
    >>> print('first scan: ', spec_data.getFirstScanNumber())
    1
    >>> print('last scan: ', spec_data.getLastScanNumber())
    22

Get plottable data from scan number 10:

    >>> from spec2nexus import spec
    >>> spec_data = spec.SpecDataFile('path/to/my/spec_data.dat')
    >>> scan10 = spec_data.getScan(10)
    >>> x_label = scan10.L[0]
    >>> y_label = scan10.L[-1]
    >>> x_data = scan10.data[x_label]
    >>> y_data = scan10.data[y_label]


Try to read a file that does not exist:

    >>> spec_data = spec.SpecDataFile('missing_file')
    Traceback (most recent call last):
      ...
    spec.SpecDataFileNotFound: file does not exist: missing_file

.. rubric::  Classes and Methods

"""

from collections import OrderedDict
import os
import time
from . import plugin


UNRECOGNIZED_KEY = "unrecognized_control_line"
MCA_DATA_KEY = "_mca_"


class SpecDataFileNotFound(IOError):
    """data file was not found"""


class SpecDataFileCouldNotOpen(IOError):
    """data file could not be opened"""


class NotASpecDataFile(Exception):
    """content of file is not SPEC data (first line must start with ``#F``)"""


class DuplicateSpecScanNumber(Exception):
    """multiple use of scan number in a single SPEC data file"""


class UnknownSpecFilePart(Exception):
    """unknown part in a single SPEC data file"""


def is_spec_file(filename):
    """
    test if a given file name is a SPEC data file

    :param str filename: path/to/possible/spec/data.file

    *filename* is a SPEC file if it contains at least one #S control line
    """
    if not os.path.exists(filename) or not os.path.isfile(filename):
        return False
    try:
        with open(filename, "r") as fp:
            for line in fp.readlines():
                if line.startswith("#S "):
                    return True
    except Exception:
        pass
    return False


def is_spec_file_with_header(filename):
    """
    test if a given file name is a SPEC data file

    :param str filename: path/to/possible/spec/data.file

    *filename* is a SPEC file only if the file starts [#]_
    with these control lines in order:

    * #F    - original filename
    * #E    - the UNIX epoch (seconds from 00:00 GMT 1/1/70)
    * #D    - current date and time in UNIX format
    * #C    - comment line (the first one provides the filename again and the user name)

    such as::

        #F LNO_LAO
        #E 1276730676
        #D Wed Jun 16 18:24:36 2010
        #C LNO_LAO  User = epix33bm

    .. [#] SPEC manual, *Standard Data File Format*, http://www.certif.com/spec_manual/user_1_4_1.html
    """
    if not os.path.exists(filename) or not os.path.isfile(filename):
        return False
    expected_controls = ("#F ", "#E ", "#D ", "#C ")
    try:
        lines = open(filename).readlines()[: len(expected_controls)]
    except UnicodeDecodeError:
        return False
    if len(lines) != len(expected_controls):
        return False
    for expected, line in zip(expected_controls, lines):
        if not line.startswith(expected):
            return False
    return True


# -------------------------------------------------------------------------------------------


class SpecDataFile(object):

    """
    contents of a SPEC data file

    .. autosummary::

        ~dissect_file
        ~getFirstScanNumber
        ~getLastScanNumber
        ~getMaxScanNumber
        ~getMinScanNumber
        ~getScan
        ~getScanCommands
        ~getScanNumbers
        ~getScanNumbersChronological
        ~read
        ~refresh
        ~update_available

    """

    fileName = ""
    parts = ""
    headers = []
    scans = {}
    readOK = -1

    def __init__(self, filename):
        self.fileName = None
        self.headers = []
        self.scans = OrderedDict()
        self.readOK = -1
        self.last_scan = None
        self.mtime = 0
        self.filesize = 0

        if filename is not None:
            if not os.path.exists(filename):
                raise SpecDataFileNotFound(
                    "file does not exist: " + str(filename)
                )
            if not is_spec_file(filename):
                raise NotASpecDataFile(
                    "not a SPEC data file: " + str(filename)
                )
            self.fileName = filename

            self.read()

    def __str__(self):
        return self.fileName or "None"

    @property
    def update_available(self):
        """
        Has the file been updated since the last time it was read?

        Reference file modification time is stored *after*
        file is read in :meth:`read()` method.

        EXAMPLE USAGE

        Open the SPEC data file (example):

            sdf = spec.SpecDataFile(filename)

        then, monitor (continuing example):

            if sdf.update_available:
                myLastScan = sdf.last_scan
                sdf.read()
                plot_scan_and_newer(myLastScan)    # new method
                myLastScan = sdf.last_scan

        .. note: The previous last_scan is reprocessed since
           that scan may not have been complete when the file
           was read() previously.

        """
        same_mtime = self.mtime == os.path.getmtime(self.fileName)
        same_size = self.filesize == os.path.getsize(self.fileName)
        identical = same_mtime and same_size
        return not identical

    def refresh(self):
        """
        update (refresh) the content if the file is updated

        returns previous last_scan or None if file not updated

        .. caution:  previous last_scan must be re-created if updated

           After calling :meth:`refresh()`, any client
           with an object of the previous last scan
           should get a new object with the update data.

           EXAMPLE::

               scan = sdf.getScan(42)
               checkpoint = sdf.refresh()
               if checkpoint is not None:
                   scan = sdf.getScan(checkpoint)    # get updates

        """
        if self.update_available:
            previous_scan = self.last_scan

            self.read()
            return previous_scan
        return None

    def _read_file_(self, spec_file_name):
        """Reads a spec data file"""
        if not os.path.exists(spec_file_name):
            raise SpecDataFileNotFound(
                "file does not exist: " + str(spec_file_name)
            )
        try:
            with open(spec_file_name, "r") as fp:
                buf = fp.read()
        except IOError:
            msg = "Could not open spec file: " + str(spec_file_name)
            raise SpecDataFileCouldNotOpen(msg)
        if not is_spec_file(spec_file_name):
            msg = "Not a spec data file: " + str(spec_file_name)
            raise NotASpecDataFile(msg)

        # caution: some files may have EOL = \r\n
        # convert all '\r\n' to '\n', then all '\r' to '\n'

        return buf.replace("\r\n", "\n").replace("\r", "\n")

    def dissect_file(self):
        """
        divide (SPEC data file text) buffer into sections

        internal: A *block* starts with either #F | #E | #S

        RETURNS

        [block]
            list of blocks where each block is one or more lines of
            text with one of the above control lines at its start

        """
        buf = self._read_file_(self.fileName).splitlines()

        sections, block = [], []

        for _line_num, text in enumerate(buf):
            if len(text.strip()) > 0:
                f = text.split()[0]
                if len(f) == 2 and f in ("#E", "#F", "#S"):
                    if len(block) > 0:
                        sections.append("\n".join(block))
                    block = []
            block.append(text)

        if len(block) > 0:
            sections.append("\n".join(block))
        return sections

    def read(self):
        """Reads and parses a spec data file"""
        manager = plugin.get_plugin_manager()
        sections = self.dissect_file()
        for block in sections:
            if len(block) == 0:
                continue
            key = manager.getKey(block.splitlines()[0])
            manager.process(key, block, self)

            if key == "#S":
                scan = list(self.scans.values())[-1]
                for line in scan.raw.splitlines()[1:]:
                    if len(line) > 0:
                        key = line.split()[0]
                        if key in ("#D",):
                            manager.process(key, line, scan)
                            break

        # fix any missing parts
        if not hasattr(self, "specFile"):
            self.specFile = self.fileName

        self.last_scan = self.getLastScanNumber()
        self.filesize = os.path.getsize(self.fileName)
        self.mtime = os.path.getmtime(self.fileName)

    def getScan(self, scan_number=0):
        """return the scan number indicated, None if not found"""
        if int(float(scan_number)) < 1:
            # relative scan reference
            scanlist = self.getScanNumbers()
            scan_number = list(scanlist)[int(scan_number)]
        scan_number = str(scan_number)
        if scan_number in self.scans:
            return self.scans[scan_number]
        return None

    def getScanNumbers(self):
        """return a list of all scan numbers sorted by scan number"""
        keys = self.scans.keys()
        try:
            r = sorted(keys, key=int)
        except ValueError as _exc:
            r = sorted(keys, key=float)
        return r

    def getScanNumbersChronological(self):
        """return a list of all scan numbers sorted by date"""

        def byDate_key(scan):
            return time.strptime(scan.date)

        scans = sorted(self.scans.values(), key=byDate_key)
        return [_.scanNum for _ in scans]

    def getMinScanNumber(self):
        """return the lowest numbered scan"""
        return self.getScanNumbers()[0]

    def getMaxScanNumber(self):
        """return the highest numbered scan"""
        return self.getScanNumbers()[-1]

    def getFirstScanNumber(self):
        """return the first scan"""
        return self.getScanNumbersChronological()[0]

    def getLastScanNumber(self):
        """return the last scan"""
        return self.getScanNumbersChronological()[-1]

    def getScanCommands(self, scan_list=None):
        """return all the scan commands as a list, with scan number"""
        scan_list = scan_list or self.getScanNumbers()
        commands = []
        for key in scan_list:
            scan = self.getScan(key)
            if isinstance(scan, SpecDataFileScan):
                commands.append("#S " + str(key) + " " + scan.scanCmd)
        return commands


# -------------------------------------------------------------------------------------------


class SpecDataFileHeader(object):
    """
    contents of a spec data file header (#E) section

    .. autosummary::

        ~interpret
        ~addPostProcessor
        ~addH5writer
        ~getLatestScan

    """

    def __init__(self, buf, parent=None):
        # ----------- initialize the instance variables
        self.parent = parent  # instance of SpecDataFile
        self.comments = []
        self.date = ""
        self.epoch = 0
        if parent is None:
            self.file = None
        else:
            self.file = parent.fileName
        self.H = []
        self.O = []
        self.raw = buf
        self.postprocessors = {}
        self.h5writers = {}

    def interpret(self):
        """ interpret the supplied buffer with the spec data file header"""
        manager = plugin.get_plugin_manager()
        for _i, line in enumerate(self.raw.splitlines(), start=1):
            if len(line) == 0:
                continue  # ignore blank lines
            key = manager.getKey(line)
            if key is None:
                # log message instead of raise exception
                # https://github.com/prjemian/spec2nexus/issues/57
                # raise UnknownSpecFilePart("line %d: unknown header line: %s" % (_i, line))
                key = UNRECOGNIZED_KEY
                manager.process(key, line, self)
            elif key == "#E":
                pass  # avoid recursion
            else:
                # most of the work is done here
                manager.process(key, line, self)

        # call any post-processing hook functions from the plugins
        for func in self.postprocessors.values():
            func(self)

    def addPostProcessor(self, label, func):
        """
        add a function to be processed after interpreting all lines from a header

        :param str label: unique label by which this postprocessor will be known
        :param obj func: function reference of postprocessor

        The postprocessors will be called at the end of header interpretation.
        """
        if label not in self.postprocessors:
            self.postprocessors[label] = func

    def addH5writer(self, label, func):
        """
        add a function to be processed when writing the scan header

        :param str label: unique label by which this writer will be known
        :param obj func: function reference of writer

        The writers will be called when the HDF5 file is to be written.
        """
        if label not in self.h5writers:
            self.h5writers[label] = func

    def getLatestScan(self):
        return list(self.parent.scans.values())[-1]


# -------------------------------------------------------------------------------------------


class SpecDataFileScan(object):
    """
    contents of a spec data file scan (#S) section

    .. autosummary::

        ~get_macro_name
        ~interpret
        ~add_interpreter_comment
        ~get_interpreter_comments
        ~addPostProcessor
        ~addH5writer

    """

    def __init__(self, header, buf, parent=None):
        self.parent = parent  # instance of SpecDataFile
        self.comments = []
        self.data = {}
        self.data_lines = []
        self.date = ""
        self.G = {}
        self.header = header  # index number of relevant #F section previously interpreted
        self.L = []
        self.M = ""
        self.positioner = {}
        self.N = -1
        self.P = []
        self.Q = ""
        self.raw = buf
        self.S = ""
        self.scanNum = -1
        self.scanCmd = ""
        self._interpreter_comments_ = []
        if parent is not None:
            # avoid changing the interface for clients
            if isinstance(parent, SpecDataFile):
                self.specFile = parent.fileName
            elif isinstance(parent, SpecDataFileHeader):
                self.specFile = parent.parent.fileName
        elif self.header is not None:
            if self.header.parent is not None:
                self.specFile = self.header.parent.fileName
        else:
            self.specFile = None
        self.T = ""
        self.V = []
        self.column_first = ""
        self.column_last = ""
        self.postprocessors = {}
        self.h5writers = {}

        # the attributes defined in PluginManager().lazy_attributes
        # are set only after a call to self.interpret()
        # That call is triggered on the first call for any of these attributes.
        self.__lazy_interpret__ = True
        self.__interpreted__ = False

    def __str__(self):
        return self.S

    def __getattribute__(self, attr):
        manager = plugin.get_plugin_manager()
        if attr in manager.lazy_attributes:
            if self.__lazy_interpret__:
                self.interpret()
        return object.__getattribute__(self, attr)

    def get_macro_name(self):
        """
        name of the SPEC macro used for this scan
        """
        return self.scanCmd.split()[0]

    def interpret(self):
        """interpret the supplied buffer with the spec scan data"""
        manager = plugin.get_plugin_manager()
        if self.__interpreted__:  # do not do this twice
            return
        self.__lazy_interpret__ = False  # set now to avoid recursion
        lines = self.raw.replace("\\\n", " ").splitlines()
        for _i, line in enumerate(lines, start=1):
            if len(line) == 0:
                continue  # ignore blank lines
            key = manager.getKey(line.lstrip())
            if key is None:
                # __s__ = '<' + line + '>'
                # _msg = "scan %s, line %d: unknown key, ignored text: %s" % (str(self.scanNum), _i, line)
                # raise UnknownSpecFilePart(_msg)
                # log message instead of raise exception
                # https://github.com/prjemian/spec2nexus/issues/57
                key = UNRECOGNIZED_KEY
                manager.process(key, line, self)
            elif key != "#S":  # avoid recursion
                # most of the work is done here
                manager.process(key, line, self)

        # call any post-processing hook functions from the plugins
        for func in self.postprocessors.values():
            func(self)

        self.__interpreted__ = True

    def add_interpreter_comment(self, comment):
        """
        allow the interpreter to communicate information to the caller

        see issue #66: https://github.com/prjemian/spec2nexus/issues/66
        """
        self._interpreter_comments_.append(comment)

    def get_interpreter_comments(self):
        """
        return the list of comments

        see issue #66: https://github.com/prjemian/spec2nexus/issues/66
        """
        return self._interpreter_comments_

    def addPostProcessor(self, label, func):
        """
        add a function to be processed after interpreting all lines from a scan

        :param str label: unique label by which this postprocessor will be known
        :param obj func: function reference of postprocessor

        The postprocessors will be called at the end of scan data interpretation.
        """
        if label not in self.postprocessors:
            self.postprocessors[label] = func

    def addH5writer(self, label, func):
        """
        add a function to be processed when writing the scan data

        :param str label: unique label by which this writer will be known
        :param obj func: function reference of writer

        The writers will be called when the HDF5 file is to be written.
        """
        if label not in self.h5writers:
            self.h5writers[label] = func

    def _interpret_data_row(self, row_text):
        buf = {}
        for col, val in enumerate(row_text.split()):
            buf[self.L[col]] = float(val)
        return buf

    def _unique_key(self, label, keylist):
        """ensure that label is not yet existing in keylist"""
        i = 0
        key = label
        while key in keylist:
            i += 1
            key = label + "_" + str(i)
            if i == 1000:
                raise RuntimeError(
                    "cannot make unique key for duplicated column label!"
                )
        return key



# -*- coding: utf-8 -*-

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     prjemian@gmail.com
# :copyright: (c) 2014-2020, Pete R. Jemian
#
# Distributed under the terms of the Creative Commons Attribution 4.0 International Public License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------


"""
(internal library) common methods used in **spec2nexus** modules

.. autosummary::

    ~clean_name
    ~iso8601
    ~strip_first_word
    ~sanitize_name
    ~reshape_data

"""

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     prjemian@gmail.com
# :copyright: (c) 2014-2017, Pete R. Jemian
#
# Distributed under the terms of the Creative Commons Attribution 4.0 International Public License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------

import logging
import numpy
import re
import time


logger = logging.getLogger(__name__)


def clean_name(key):
    """
    create a name that is allowed by both HDF5 and NeXus rules

    :param str key: identifying string from SPEC data file

    :see: http://download.nexusformat.org/doc/html/datarules.html

    The "sanitized" name fits this regexp::

        [A-Za-z_][\w_]*

    An easier expression might be:  ``[\w_]*`` but this will not pass
    the rule that valid NeXus group or field names cannot start with a digit.
    """
    replacement = "_"
    noncompliance = r"[^\w_]"
    txt = replacement.join(
        re.split(noncompliance, key)
    )  # replace ALL non-compliances with '_'
    if txt[0].isdigit():
        txt = replacement + txt  # can't start with a digit
    return txt


def iso8601(date):
    """
    convert SPEC time (example: Wed Nov 03 13:39:34 2010) into ISO8601 string

    :param str date: time string from SPEC data file

    **Example**

    :SPEC:    Wed Nov 03 13:39:34 2010
    :ISO8601: 2010-11-03T13:39:34
    :SPOCK:   09/15/17 04:39:10
    :ISO8601: 2017-09-15T04:39:10
    """
    spec_fmt = "%a %b %d %H:%M:%S %Y"
    try:
        t_obj = time.strptime(date, spec_fmt)
    except ValueError as ex:
        spock_fmt = "%m/%d/%y %H:%M:%S"
        try:
            t_obj = time.strptime(date, spock_fmt)
        except ValueError as ex:
            raise ex
    iso_fmt = "%Y-%m-%dT%H:%M:%S"
    iso = time.strftime(iso_fmt, t_obj)
    return iso


def strip_first_word(line):
    """return everything after the first space on the line from the spec data file"""
    pos = line.find(" ")
    val = line[pos:]
    return val.strip()


def split_column_labels(text):
    """SPEC labels may contain one space"""
    return re.split("  +", text.replace("\t", "  "))


def sanitize_name(group, key):  # for legacy support only
    """make name that is allowed by HDF5 and NeXus rules

    :note: **deprecated**  use :func:`clean_name` instead (``group`` is never used)
    :param str group: unused
    :param str key: identifying string from SPEC data file

    :see: http://download.nexusformat.org/doc/html/datarules.html

    sanitized name fits this regexp::

        [A-Za-z_][\w_]*

    An easier expression might be:  ``[\w_]*`` but this will not pass
    the rule that valid names cannot start with a digit.
    """
    return clean_name(key)


def reshape_data(scan_data, scan_shape):
    """
    Shape scan data from raw to different dimensionality

    Some SPEC macros collect data in a mesh or grid yet
    report the data as a 1-D sequence of observations.
    For further processing (such as plotting), the scan data
    needs to be reshaped according to its intended dimensionality.

    modified from nexpy.readers.readspec.reshape_data
    """
    scan_size = numpy.prod(scan_shape)
    if scan_data.size == scan_size:
        data = scan_data
    elif scan_data.size < scan_size:
        data = numpy.empty(scan_size)
        data.fill(numpy.NaN)  # pad data with NaN
        data[0 : scan_data.size] = scan_data.ravel()  # flatten & insert
    else:
        data = scan_data.ravel()  # flatten
        data = data[0:scan_size]  # truncate extra data
    return data.reshape(scan_shape)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     prjemian@gmail.com
# :copyright: (c) 2014-2020, Pete R. Jemian
#
# Distributed under the terms of the Creative Commons Attribution 4.0 International Public License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------


"""Converts SPEC data files and scans into NeXus HDF5 files"""

# this is the main code for the *spec2nexus* application

__url__ = "http://spec2nexus.readthedocs.org/en/latest/spec2nexus.html"

import os
import sys

if __name__ == "__main__":
    # put us on the path for developers
    path = os.path.join("..", os.path.dirname(__file__))
    sys.path.insert(0, os.path.abspath(path))

from . import spec
from . import writer


hdf5_extension = ".hdf5"


# -------------------------------------------------------------------------------------------


REPORTING_QUIET = "quiet"
REPORTING_STANDARD = "standard"
REPORTING_VERBOSE = "verbose"
SCAN_LIST_ALL = "all"


def get_user_parameters():
    """configure user's command line parameters from sys.argv"""
    # global hdf5_extension
    import argparse
    from spec2nexus._version import get_versions

    version = get_versions()["version"]
    doc = __doc__.strip().splitlines()[0]
    doc += "\n  URL: " + __url__
    doc += "\n  v" + version
    parser = argparse.ArgumentParser(prog="spec2nexus", description=doc)
    parser.add_argument(
        "infile", action="store", nargs="+", help="SPEC data file name(s)"
    )
    msg = "NeXus HDF5 output file extension"
    msg += ", default = %s" % hdf5_extension
    parser.add_argument(
        "-e",
        "--hdf5-extension",
        action="store",
        dest="hdf5_extension",
        help=msg,
        default=hdf5_extension,
    )
    parser.add_argument(
        "-f",
        "--force-overwrite",
        action="store_true",
        dest="force_write",
        help="overwrite output file if it exists",
        default=False,
    )
    parser.add_argument(
        "-v", "--version", action="version", version=version
    )
    msg = "specify which scans to save"
    msg += ", such as: -s all  or  -s 1  or  -s 1,2,3-5  (no spaces!)"
    msg += ", default = %s" % SCAN_LIST_ALL
    parser.add_argument(
        "-s",
        "--scan",
        nargs=1,
        # action='append',
        dest="scan_list",
        default=SCAN_LIST_ALL,
        help=msg,
    )
    #     parser.add_argument('-t',
    #                         '--tree-only',
    #                         action='store_true',
    #                         dest='tree_only',
    #                         help='print NeXus/HDF5 node tree (does not save to a file)',
    #                         default=False)

    group = parser.add_mutually_exclusive_group()
    group.set_defaults(reporting_level=REPORTING_STANDARD)
    msg = "suppress all program output (except errors)"
    msg += ", do not use with --verbose option"
    group.add_argument(
        "--quiet",
        dest="reporting_level",
        action="store_const",
        const=REPORTING_QUIET,
        help=msg,
    )
    msg = "print more program output"
    msg += ", do not use with --quiet option"
    group.add_argument(
        "--verbose",
        dest="reporting_level",
        action="store_const",
        const=REPORTING_VERBOSE,
        help=msg,
    )

    return parser.parse_args()


def parse_scan_list_spec(scan_list_spec):
    """parses the argument of the -s option, returns a scan number list"""
    # can this be simpler?
    sl = scan_list_spec[0].split(",")

    scan_list = []
    for item in sl:
        sublist = item.split("-")
        if len(sublist) == 1:
            scan_list.append(int(sublist[0]))
        elif len(sublist) == 2:
            scan_list += range(int(sublist[0]), int(sublist[1]) + 1)
        else:
            raise ValueError("improper scan list specifier: " + sublist)

    sl = []
    for item in sorted(scan_list):
        if item not in sl:
            sl.append(item)

    return sl


def pick_scans(all_scans, opt_scan_list):
    """
    edit opt_scan_list for the scans to be converted

    To be converted, a scan number must be first specified in opt_scan_list
    and then all_scans is checked to make sure that scan exists.
    The final list is returned.
    """
    if opt_scan_list == SCAN_LIST_ALL:
        scan_list = all_scans
    else:
        scan_list = map(str, opt_scan_list)
        # for item in scan_list:
        #     if str(item) not in all_scans:
        #         scan_list.remove(item)
        scan_list = [item for item in scan_list if str(item) in all_scans]
    return scan_list


def main():
    """entry point for command-line interface"""

    user_parms = get_user_parameters()

    spec_data_file_name_list = user_parms.infile

    if user_parms.scan_list != SCAN_LIST_ALL:
        user_parms.scan_list = parse_scan_list_spec(user_parms.scan_list)

    if not user_parms.hdf5_extension.startswith(os.extsep):
        user_parms.hdf5_extension = os.extsep + user_parms.hdf5_extension

    for spec_data_file_name in spec_data_file_name_list:
        if not os.path.exists(spec_data_file_name):
            msg = "File not found: " + spec_data_file_name
            print(msg)
            continue

        if user_parms.reporting_level in (
            REPORTING_STANDARD,
            REPORTING_VERBOSE,
        ):
            print("reading SPEC data file: " + spec_data_file_name)
        spec_data = spec.SpecDataFile(spec_data_file_name)

        all_scans = spec_data.getScanNumbers()
        scan_list = list(pick_scans(all_scans, user_parms.scan_list))
        if user_parms.reporting_level in (REPORTING_VERBOSE):
            print("  discovered %d scans" % len(spec_data.scans.keys()))
            print(
                "  converting scan number(s): "
                + ", ".join(map(str, scan_list))
            )

        basename = os.path.splitext(spec_data_file_name)[0]
        nexus_output_file_name = basename + user_parms.hdf5_extension
        if user_parms.force_write or not os.path.exists(
            nexus_output_file_name
        ):
            out = writer.Writer(spec_data)
            out.save(nexus_output_file_name, scan_list)
            if user_parms.reporting_level in (
                REPORTING_STANDARD,
                REPORTING_VERBOSE,
            ):
                print("wrote NeXus HDF5 file: " + nexus_output_file_name)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Writes a simple NeXus HDF5 file using h5py with links.

This example is based on ``writer_2_1`` of the NeXus Manual:
http://download.nexusformat.org/doc/html/examples/h5py/index.html
"""

from spec2nexus import eznx


HDF5_FILE = "eznx_example.hdf5"

I_v_TTH_DATA = """
17.92608    1037
17.92558    2857
17.92508    23819
17.92458    49087
17.92408    66802
17.92358    66206
17.92308    64129
17.92258    56795
17.92208    29315
17.92158    6622
17.92108    1321
"""
# ---------------------------

tthData, countsData = zip(
    *[map(float, _.split()) for _ in I_v_TTH_DATA.strip().splitlines()]
)

f = eznx.makeFile(HDF5_FILE)  # create the HDF5 NeXus file
f.attrs["default"] = "entry"

nxentry = eznx.makeGroup(f, "entry", "NXentry", default="data")
nxinstrument = eznx.makeGroup(nxentry, "instrument", "NXinstrument")
nxdetector = eznx.makeGroup(nxinstrument, "detector", "NXdetector")

tth = eznx.makeDataset(nxdetector, "two_theta", tthData, units="degrees")
counts = eznx.makeDataset(nxdetector, "counts", countsData, units="counts")

nxdata = eznx.makeGroup(
    nxentry,
    "data",
    "NXdata",
    signal=1,
    axes="two_theta",
    two_theta_indices=0,
)
eznx.makeLink(nxdetector, tth, nxdata.name + "/two_theta")
eznx.makeLink(nxdetector, counts, nxdata.name + "/counts")

f.close()  # be CERTAIN to close the file



# -*- coding: iso-8859-1 -*-

"""command-line tool to convert SPEC data files to NeXus HDF5"""

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     prjemian@gmail.com
# :copyright: (c) 2014-2020, Pete R. Jemian
#
# Distributed under the terms of the Creative Commons Attribution 4.0 International Public License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------


__author__ = "Pete R. Jemian"
__email__ = "prjemian@gmail.com"
__copyright__ = "2014-2020, Pete R. Jemian"

__package_name__ = "spec2nexus"

__license_url__ = "http://creativecommons.org/licenses/by/4.0/deed.en_US"
__license__ = "Creative Commons Attribution 4.0 International Public License (see LICENSE file)"
__description__ = (
    "Converts SPEC data files and scans into NeXus HDF5 files"
)
__author_name__ = __author__
__author_email__ = __email__
__url__ = u"http://spec2nexus.readthedocs.org"
__keywords__ = [
    "SPEC",
    "diffraction",
    "data acquisition",
    "NeXus",
    "HDF5",
    "MatPlotLib",
]

from ._requirements import learn_requirements

__install_requires__ = learn_requirements()
del learn_requirements

__classifiers__ = [
    "Development Status :: 5 - Production/Stable",
    "Environment :: Console",
    "Intended Audience :: Science/Research",
    "License :: Freely Distributable",
    "License :: Public Domain",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.5",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Topic :: Scientific/Engineering",
    "Topic :: Scientific/Engineering :: Astronomy",
    "Topic :: Scientific/Engineering :: Bio-Informatics",
    "Topic :: Scientific/Engineering :: Chemistry",
    "Topic :: Scientific/Engineering :: Information Analysis",
    "Topic :: Scientific/Engineering :: Interface Engine/Protocol Translator",
    "Topic :: Scientific/Engineering :: Mathematics",
    "Topic :: Scientific/Engineering :: Physics",
    "Topic :: Scientific/Engineering :: Visualization",
    "Topic :: Software Development",
    "Topic :: Utilities",
]

from ._version import get_versions

__version__ = get_versions()["version"]
del get_versions



# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     jemian@anl.gov
# :copyright: (c) 2017-2020, UChicago Argonne, LLC
#
# Distributed under the terms of the Creative Commons Attribution 4.0 International Public License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------

REQUIREMENTS = "requirements.txt"


def learn_requirements():
    """
    list all installation requirements

    ALL packages & version restrictions stated in requirements.txt
    """
    import os

    path = os.path.dirname(__file__)
    req_file = os.path.join(path, "..", "..", REQUIREMENTS)
    reqs = []
    if not os.path.exists(req_file):
        # not needed with installed package
        return reqs

    excludes = "versioneer coveralls coverage".split()
    with open(req_file, "r") as fp:
        buf = fp.read().strip().splitlines()
        for req in buf:
            req = req.strip()
            if (
                req != ""
                and not req.startswith("#")
                and req not in excludes
            ):
                reqs.append(req)
    return reqs



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     prjemian@gmail.com
# :copyright: (c) 2014-2020, Pete R. Jemian
#
# Distributed under the terms of the Creative Commons Attribution 4.0 International Public License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------


"""
Describe SPEC #G control lines

  #G0 ...          geometry parameters from G[] array (geo mode, sector, etc)
  #G1 ...          geometry parameters from U[] array (lattice constants, orientation reflections)
  #            G2 is unused
  #G3 ...          geometry parameters from UB[] array (orientation matrix)
  #G4 ...          geometry parameters from Q[] array (lambda, frozen angles, cut points, etc)

API

.. autosummary::

    ~Diffractometer
    ~DiffractometerGeometryCatalog
    ~split_name_variation
    ~get_geometry_catalog
    ~reset_geometry_catalog


"""

from collections import namedtuple, OrderedDict
import logging
import numpy
import os


logger = logging.getLogger(__name__)
_path = os.path.dirname(__file__)
DICT_FILE = os.path.join(_path, "diffractometer-geometries.dict")


KeyDescriptionValue = namedtuple(
    "KeyDescriptionValue", "key description value"
)

LatticeParameters = namedtuple(
    "LatticeParameters", "a b c alpha beta gamma"
)

Reflections = namedtuple("Reflections", "h k l wavelength angles")


# lattice constants, orientation reflections
# most geometries use this default schema
# at least one geometry (twoc) has overrides
DEFAULT_U = [
    ("g_aa", "a lattice constant (real space)"),
    ("g_bb", "b lattice constant (real space)"),
    ("g_cc", "c lattice constant (real space)"),
    ("g_al", "alpha lattice angle (real space)"),
    ("g_be", "beta  lattice angle (real space)"),
    ("g_ga", "gamma lattice angle (real space)"),
    ("g_aa_s", "a lattice constant (reciprocal space)"),
    ("g_bb_s", "b lattice constant (reciprocal space)"),
    ("g_cc_s", "c lattice constant (reciprocal space)"),
    ("g_al_s", "alpha lattice angle (reciprocal space)"),
    ("g_be_s", "beta  lattice angle (reciprocal space)"),
    ("g_ga_s", "gamma lattice angle (reciprocal space)"),
    ("g_h0", "H of primary reflection"),
    ("g_k0", "K of primary reflection"),
    ("g_l0", "L of primary reflection"),
    ("g_h1", "H of secondary reflection"),
    ("g_k1", "K of secondary reflection"),
    ("g_l1", "L of secondary reflection"),
    ("g_u00", "angle 0 of primary reflection"),
    ("g_u01", "angle 1 of primary reflection"),
    ("g_u02", "angle 2 of primary reflection"),
    ("g_u03", "angle 3 of primary reflection"),
    ("g_u04", "angle 4 of primary reflection"),
    ("g_u05", "angle 5 of primary reflection"),
    ("g_u10", "angle 0 of secondary reflection"),
    ("g_u11", "angle 1 of secondary reflection"),
    ("g_u12", "angle 2 of secondary reflection"),
    ("g_u13", "angle 3 of secondary reflection"),
    ("g_u14", "angle 4 of secondary reflection"),
    ("g_u15", "angle 5 of secondary reflection"),
    ("g_lambda0", "lambda when or0 was set"),
    ("g_lambda1", "lambda when or1 was set"),
    # For backward compatibility, additional angles added at end of array
    ("g_u06", "optional, angle 6 of primary reflection"),
    ("g_u16", "optional, angle 6 of secondary reflection"),
]


def split_name_variation(geo_name):
    """
    split geo_name into geometry name and variation
    """
    nm = geo_name
    try:
        nm, variant = geo_name.split(".")
    except ValueError:
        variant = None
    return nm, variant


class Diffractometer:
    """
    describe the diffractometer for the scan

    .. autosummary::

        ~parse
    """

    def __init__(self, geo_name):
        self.geometry_name_full = geo_name
        self.geometry_name, self.variant = split_name_variation(geo_name)
        self.geometry_parameters = {}  # combined #G terms
        self.lattice = None
        self.mode = None
        self.reflections = []
        self.sector = None
        self.wavelength = None

        gpar = self.geometry_parameters
        gpar["diffractometer_full"] = KeyDescriptionValue(
            "diffractometer_full",
            "name of diffractometer (and variant), deduced from scan information",
            self.geometry_name_full,
        )
        gpar["diffractometer_simple"] = KeyDescriptionValue(
            "diffractometer_simple",
            "name of diffractometer, deduced from scan information",
            self.geometry_name,
        )
        gpar["diffractometer_variant"] = KeyDescriptionValue(
            "diffractometer_variant",
            "name of diffractometer variant, deduced from scan information",
            self.variant or "",
        )

    def parse(self, scan):
        dgc = DiffractometerGeometryCatalog()
        gonio = dgc.get(self.geometry_name_full)

        G = OrderedDict()
        if "G0" in scan.G:
            g0 = list(map(float, scan.G["G0"].split()))
            if len(g0) == len(gonio["G"]):
                G = OrderedDict(
                    {
                        kd[0]: KeyDescriptionValue(kd[0], kd[1], v)
                        for v, kd in zip(g0, gonio["G"])
                    }
                )
                additions = []
                for k, kdv in G.items():
                    if k.startswith("g_sect"):
                        self.sector = int(kdv.value)
                        continue
                    if not k.startswith("g_mode"):
                        continue
                    modes = gonio["modes"]
                    if not (0 <= int(kdv.value) < len(modes)):
                        continue
                    entry = KeyDescriptionValue(
                        k + "_name",
                        "name of " + kdv.description,
                        modes[int(kdv.value)],
                    )
                    additions.append(entry)
                # must update G{} in two steps
                for entry in additions:
                    G[entry.key] = entry
                    if entry.key == "g_mode_name":
                        self.mode = entry.value
        self.geometry_parameters.update(G)

        if "G1" in scan.G:
            g1 = list(map(float, scan.G["G1"].split()))
            if len(g1) > 1:
                # at least one geometry overrides the default U[] terms
                gonio_U = gonio.get("U", list(DEFAULT_U))
                U = OrderedDict(
                    {
                        kd[0]: KeyDescriptionValue(kd[0], kd[1], v)
                        for v, kd in zip(g1, gonio_U)
                    }
                )
                self.lattice = LatticeParameters(
                    U["g_aa"].value,
                    U["g_bb"].value,
                    U["g_cc"].value,
                    U["g_al"].value,
                    U["g_be"].value,
                    U["g_ga"].value,
                )
                for ref_num in (0, 1):
                    template = "g_u%d%%d" % ref_num
                    variant = dgc.get_variant(gonio, self.variant)
                    angles = OrderedDict()
                    for i, mne in enumerate(variant["motors"]):
                        k = template % i
                        if k in U:
                            angles[mne] = U[k].value
                    ref = Reflections(
                        U["g_h%d" % ref_num].value,
                        U["g_k%d" % ref_num].value,
                        U["g_l%d" % ref_num].value,
                        U["g_lambda%d" % ref_num].value,
                        angles,
                    )
                    self.reflections.append(ref)

                self.geometry_parameters.update(U)

        if "G3" in scan.G:
            g3 = list(map(float, scan.G["G3"].split()))
            if len(g3) == 9:
                UB = numpy.array(g3).reshape((3, 3))
                self.geometry_parameters[
                    "ub_matrix"
                ] = KeyDescriptionValue("ub_matrix", "UB[] matrix", UB)

        Q = OrderedDict()
        if "G4" in scan.G:
            g4 = list(map(float, scan.G["G4"].split()))
            if len(g4) == len(gonio["Q"]):
                Q = OrderedDict(
                    {
                        kd[0]: KeyDescriptionValue(kd[0], kd[1], v)
                        for v, kd in zip(g4, gonio["Q"])
                    }
                )
                if "LAMBDA" in Q:
                    self.wavelength = Q["LAMBDA"].value
        self.geometry_parameters.update(Q)


_geometry_catalog = (
    None  # singleton reference to DiffractometerGeometryCatalog
)


def get_geometry_catalog():
    global _geometry_catalog
    if _geometry_catalog is None:
        _geometry_catalog = 0
        _geometry_catalog = DiffractometerGeometryCatalog()
    return _geometry_catalog


def reset_geometry_catalog():
    global _geometry_catalog
    _geometry_catalog = None


class DiffractometerGeometryCatalog:
    """
    catalog of the diffractometer geometries known to SPEC

    .. autosummary::

        ~geometries
        ~get
        ~get_default_geometry
        ~has_geometry
        ~match
    """

    db = {}

    def __init__(self):
        if _geometry_catalog is None:
            msg = "Do not create DiffractometerGeometryCatalog()."
            msg += "  Instead, call: get_geometry_catalog()"
            raise RuntimeError(msg)

        with open(DICT_FILE, "r") as fp:
            self.db = eval(fp.read())

        self._default_geometry = self.db[0]
        self._geometries_simple = None
        self._geometries_full = None

    def __str__(self):
        v = "number=" + str(len(self.db))
        s = "DiffractometerGeometryCatalog(%s)" % v
        return s

    def geometries(self, variations=False):
        """
        list known geometries

        PARAMETERS

        variations : bool
            If True, also list known variations
        """
        if (
            self._geometries_simple is None
            and self._geometries_full is None
        ):
            # optimize with a cache, construct only once
            self._geometries_simple, self._geometries_full = [], []
            for geometry in self.db:
                nm = geometry["name"]
                self._geometries_full += [
                    "%s.%s" % (nm, v["name"])
                    for v in geometry["variations"]
                ]
                self._geometries_simple.append(nm)

        choices = {
            True: self._geometries_full,
            False: self._geometries_simple,
        }
        return choices[variations]

    def get(self, geo_name, default=None):
        """
        return dictionary for diffractometer geometry ``geo_name``
        """
        nm = split_name_variation(geo_name)[0]
        geometries = self.geometries()
        if nm in geometries:
            return self.db[geometries.index(nm)]
        return default

    def get_default_geometry(self):
        " "
        return self._default_geometry

    def get_variant(self, geometry, variant_name):
        " "
        for v in geometry["variations"]:
            if v["name"] == variant_name:
                return v
        return geometry["variations"][0]

    def has_geometry(self, geo_name):
        """
        Is the ``geo_name`` geometry defined?  True or False
        """
        nm, variant = split_name_variation(geo_name)
        if variant is None:
            return nm in self.geometries()
        else:
            return geo_name in self.geometries(True)

    def _get_scan_positioners_(self, scan):
        scan_positioners = []
        if hasattr(scan.header, "o"):  # prefer mnemonics
            for row in scan.header.o:
                scan_positioners += row
        elif hasattr(scan.header, "O"):
            for row in scan.header.O:
                scan_positioners += [k.lower() for k in row]
        else:
            scan_positioners = [k.lower() for k in scan.positioner.keys()]
        if len(scan_positioners) > 0:
            scan_positioners_new = []
            for k in scan_positioners:
                if k in ("2-theta", "two theta") or k.endswith("tth"):
                    k = "tth"
                elif k in ("theta",):
                    k = "th"
                elif k in ("delta",):
                    k = "del"
                elif k in ("gamma",):
                    k = "gam"
                scan_positioners_new.append(k)
            scan_positioners = scan_positioners_new

        return scan_positioners

    def match(self, scan):
        """
        find the ``geo_name`` geometry that matches the ``scan``

        If there is more than one matching geometry, pick the first one.
        """
        match = []

        scan_positioners = self._get_scan_positioners_(scan)

        if hasattr(scan, "G"):
            scan_G0 = scan.G.get("G0", "0").split()
            if scan_G0 == [
                "0",
            ]:
                scan_G0 = []

            if "G4" in scan.G:
                scan_G4 = scan.G["G4"].split()
                if scan_G4 == [
                    "0",
                ]:
                    scan_G4 = []
            else:
                scan_G4 = None

        else:
            scan_G0, scan_G4 = None, None

        for geometry in self.db:
            geo_name = geometry["name"]
            if len(scan_G0) != len(geometry["G"]):
                logger.debug("#G0 G[] did not match %s", geo_name)
                continue
            if scan_G4 is not None and len(scan_G4) != len(geometry["Q"]):
                logger.debug("#G4 Q[] did not match %s", geo_name)
                continue

            for variant in geometry["variations"]:
                var_name = variant["name"]
                n_motors = len(variant["motors"])
                if scan_positioners[:n_motors] != variant["motors"]:
                    logger.debug("motors did not match %s", geo_name)
                    continue

                others_match = True
                for mne in variant["other-motors"]:
                    if mne not in scan.positioner:
                        others_match = False
                        break
                if not others_match:
                    logger.debug(
                        "other-motors did not match %s (mne=%s)",
                        geo_name,
                        mne,
                    )
                    continue

                match.append("%s.%s" % (geo_name, var_name))

        if len(match) > 1:
            msg = "scan geometry match is not unique!"
            msg += "  Picking the first one from: " + str(match)
            msg += ", file: " + scan.specFile
            msg += ", scan: " + str(scan)
            logger.debug(msg)
        elif len(match) == 0:
            match = [self.get_default_geometry()["name"]]

        return match[0]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     prjemian@gmail.com
# :copyright: (c) 2014-2020, Pete R. Jemian
#
# Distributed under the terms of the Creative Commons Attribution 4.0 International Public License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------

"""
**#MD** : Bluesky metadata from apstools SpecWriterCallback.

EXAMPLE::

    #MD APSTOOLS_VERSION = 1.1.0
    #MD BLUESKY_VERSION = 1.5.2
    #MD EPICS_CA_MAX_ARRAY_BYTES = 1280000
    #MD EPICS_HOST_ARCH = linux-x86_64
    #MD OPHYD_VERSION = 1.3.2
    #MD beamline_id = APS USAXS 9-ID-C
    #MD datetime = 2019-04-19 10:04:44.400750
    #MD login_id = usaxs@usaxscontrol.xray.aps.anl.gov
    #MD pid = 27062
    #MD proposal_id = testing Bluesky installation
    #MD purpose = tuner
    #MD tune_md = {'width': -0.004, 'initial_position': 8.824885, 'time_iso8601': '2019-04-19 10:04:44.402643'}
    #MD tune_parameters = {'num': 31, 'width': -0.004, 'initial_position': 8.824885, 'peak_choice': 'com', 'x_axis': 'm_stage_r', 'y_axis': 'I0_USAXS'}
"""

from collections import OrderedDict
import six

from .. import eznx
from ..plugin import AutoRegister, ControlLineHandler


@six.add_metaclass(AutoRegister)
class MD_apstools(ControlLineHandler):

    """**#MD** -- Bluesky metadata from apstools SpecWriterCallback"""

    key = r"#MD\w*"
    scan_attributes_defined = ["MD"]

    def process(self, text, scan, *args, **kws):
        """read #MD lines from SPEC data file"""
        if not hasattr(scan, "MD"):
            scan.MD = OrderedDict()

        p = text.find("=")
        if p > len("# MD "):
            # f"#MD {key} = {value}"
            key = text.split()[1]
            value = text[p + 1 :].strip()
        else:
            # badly-formed #MD control line
            key = "MD_line_%d" % (len(scan.MD) + 1)
            value = text.strip()
        scan.MD[key] = value
        scan.addH5writer(self.key, self.writer)

    def writer(self, h5parent, writer, scan, nxclass=None, *args, **kws):
        """Describe how to store this data in an HDF5 NeXus file"""
        nxclass = "NXcollection"
        if hasattr(scan, "MD") and len(scan.MD) > 0:
            desc = "Bluesky metadata (as written by apstools.SpecWriterCallback)"
            group = eznx.makeGroup(
                h5parent, "bluesky_metadata", nxclass, description=desc
            )
            writer.save_dict(group, scan.MD)
